"""
Evaluation using UnconditionalEvalPipeline / ConsistencyEvalPipeline for training set structures.
"""

import os
from typing import List, Optional, Tuple, Union
from pathlib import Path
from tqdm import tqdm
from copy import deepcopy
import pickle
import argparse

import numpy as np
import rdkit
from rdkit import Chem
from rdkit.Chem import AllChem
import rdkit.Chem.rdDetermineBonds

from shepherd_score.score.constants import ALPHA

from shepherd_score.conformer_generation import embed_conformer_from_smiles, single_point_xtb_from_xyz
from shepherd_score.container import Molecule

from shepherd_score.evaluations.evaluate import ConsistencyEvalPipeline, resample_surf_scores, get_mol_from_atom_pos


if 'TMPDIR' in os.environ:
    TMPDIR = Path(os.environ['TMPDIR'])
else:
    TMPDIR = Path('./')


def run_consistency_evaluation(load_dir: str,
                               condition: str,
                               solvent: Optional[str] = None,
                               training_molblock_charges=Union[List[Tuple], None],
                               num_processes: int = 4,
                               probe_radius: float = 0.6,
                               ) -> None:
    """
    Generate a benchmarking set of molecules -- both MMFF and its corresponding xtb relaxed
    structures.

    Arguments
    ---------
    load_dir : path to dir containing "samples.pickle"
    condition : str (x1x2, x1x3, x1x4)
    """
    load_dir_path = Path(load_dir)
    if not load_dir_path.is_dir():
        raise ValueError(f'Provided path is not a directory: {load_dir_path}')

    save_file_dir = load_dir_path / 'consis_evals'
    save_file_dir.mkdir(parents=True, exist_ok=True)

    if condition in ('x1x2', 'x1x3', 'x1x4'):
        load_file_path = load_dir_path / 'samples.pickle'
        save_file_path = save_file_dir / 'consis_eval'
    else:
        load_file_path = load_dir_path / 'samples_noisy.pickle'
        save_file_path = save_file_dir / 'consis_eval'
    print(f'Saving to {save_file_path}')

    with open(load_file_path, 'rb') as f:
        samples = pickle.load(f)
    print(f'Samples loaded from {load_file_path}')


    ls_atoms_pos = []
    if condition == 'x1x2' or condition == 'x1x3' or condition == 'x1x3x4':
        ls_surf_points = []
    if condition == 'x1x3' or condition == 'x1x3x4':
        ls_surf_esp = []
    if condition == 'x1x4' or condition == 'x1x3x4':
        ls_pharm_feats = []

    for i in range(len(samples)):
        ls_atoms_pos.append(
            (samples[i]['x1']['atoms'], samples[i]['x1']['positions'])
        )
        if condition == 'x1x2':
            ls_surf_points.append(
                samples[i]['x2']['positions']
            )
        if condition == 'x1x3' or condition == 'x1x3x4':
            ls_surf_points.append(
                samples[i]['x3']['positions']
            )
            ls_surf_esp.append(
                samples[i]['x3']['charges']
            )
        if condition == 'x1x4' or condition == 'x1x3x4':
            ls_pharm_feats.append(
                (samples[i]['x4']['types'], samples[i]['x4']['positions'], samples[i]['x4']['directions'])
            )

    if condition == 'x1x3x4':
        assert isinstance(ls_surf_points, list) and isinstance(ls_surf_esp, list) and isinstance(ls_pharm_feats, list)
    elif condition == 'x1x2':
        assert isinstance(ls_surf_points, list)
        ls_surf_esp = None
        ls_pharm_feats = None
    elif condition == 'x1x3':
        assert isinstance(ls_surf_points, list) and isinstance(ls_surf_esp, list)
        ls_pharm_feats = None
    elif condition == 'x1x4':
        assert isinstance(ls_pharm_feats, list)
        ls_surf_points = None
        ls_surf_esp = None

    print(f'Initializing ConsistencyEvalPipeline for {condition}.')

    # Initialize evaluation
    consis_pipeline = ConsistencyEvalPipeline(
        generated_mols=ls_atoms_pos,
        generated_surf_points=ls_surf_points,
        generated_surf_esp=ls_surf_esp,
        generated_pharm_feats=ls_pharm_feats,
        probe_radius=probe_radius,
        pharm_multi_vector=False,
        solvent=solvent,
        random_molblock_charges=training_molblock_charges
    )

    # Run evaluation
    print('Running evaluation...')
    consis_pipeline.evaluate(num_processes=num_processes, verbose=True)
    print('Finished evaluation.')

    # Save values to numpy object
    save_global_series, save_local_df = consis_pipeline.to_pandas()

    save_file_global = save_file_dir / 'consis_eval_global.pkl'
    save_global_series.to_pickle(save_file_global)
    save_file_local = save_file_dir / 'consis_eval_local.pkl'
    save_local_df.to_pickle(save_file_local)

    print(f'Finished consistency evaluation!\nSaved global attributes to: {save_file_global}')
    print(f'Saved local attributes to: {save_file_local}')


if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--load-dir', type=str, required=True,
                        help='Path to directory to save files to.')
    parser.add_argument('--training-data', type=str, required=True,
                        help="Path to GDB training data file to randomly sample and compare scoring functions")
    parser.add_argument('--task-id', type=str, help='Task ID.')
    args = parser.parse_args()
    print(args)

    load_dir = str(args.load_dir)
    my_task_id = int(args.task_id) if args.task_id != '' else None
    training_data_file = Path(str(args.training_data))
    assert training_data_file.is_file()

    # settings for ShEPhERD generated molecules
    num_surf_points = 75 # number used to sample
    probe_radius = 0.6
    if 'moses' in load_dir:
        solvent = 'water'
        model = 'x1x3x4'
        num_processes = 16
        print('Doing Moses - aq')
    else:
        solvent = None # evaluating on GDB unconditional
        print('Doing GDB-17')

        my_task_id = int(args.task_id)

        model_types = ('x1x2', 'x1x3', 'x1x4')
        model = model_types[my_task_id]
        num_processes = 16

    load_dir_path = Path(load_dir) / model
    assert load_dir_path.is_dir()

    if 'moses' in load_dir:
        training_molblock_charges = None
    else:
        #load training data for random alignments / evaluations
        with open(training_data_file, 'rb') as f:
            training_molblock_charges = pickle.load(f)

    run_consistency_evaluation(
        load_dir=load_dir_path,
        condition=model,
        solvent=solvent,
        training_molblock_charges=training_molblock_charges,
        num_processes=num_processes,
        probe_radius=probe_radius
    )



"""
Functions used for SE(3) transformations. (Jax implementation).

Namely, converting quaternions to rotation matrices, getting an SE(3) transform from SE(3)
parameters, and applying the SE(3) transformation on a set of points.

Credit to Lewis J. Martin as this was adapted from
https://github.com/ljmartin/align/blob/main/0.2%20aligning%20principal%20moments%20of%20inertia.ipynb
and PyTorch's implementations.
"""
import jax.numpy as jnp
from jax import Array

def quaternions_to_rotation_matrix_jax(quaternions: Array) -> Array:
    """
    Converts quaternion to a rotation matrix. Jax implementation
    Adapted from PyTorch3D:
    https://pytorch3d.readthedocs.io/en/latest/_modules/pytorch3d/transforms/rotation_conversions.html#quaternion_to_matrix

    Parameters
    ----------
    quaternions : Array (4,)
        Quaternion parameters in (r,i,j,k) order.
        set.

    Returns
    -------
    rotation_matrix : Array (3,3)
        Rotation matrix converted from quaternion.
    """
    # Single instance
    r, i, j, k = quaternions
    two_s = 2. / (quaternions * quaternions).sum()

    rotation_matrix = jnp.stack(
        (
            1 - two_s * (j * j + k * k),
            two_s * (i * j - k * r),
            two_s * (i * k + j * r),
            two_s * (i * j + k * r),
            1 - two_s * (i * i + k * k),
            two_s * (j * k - i * r),
            two_s * (i * k - j * r),
            two_s * (j * k + i * r),
            1 - two_s * (i * i + j * j),
        ),
        -1,
    )
    return rotation_matrix.reshape((3, 3))


def get_SE3_transform_jax(se3_params: Array) -> Array:
    """
    Constructs an SE(3) transformtion matrix from parameters. Jax implementation

    Parameters
    ----------
    se3_params : Array (7,)
        Parameters for SE(3) transformation.
        The first 4 values in the last dimension are quaternions of form (r,i,j,k)
        and the last 3 values of the last dimension are the translations in (x,y,z).

    Returns
    -------
    se3_matrix : Array (4, 4)
        se3_params converted to a 4x4 SE(3) transformation matrix.
    """
    # Extract quaternion and translation parameters
    quaternion_params = se3_params[:4]
    translation_params = se3_params[4:]

    # Normalize quaternion to ensure unit length
    quaternion_params = quaternion_params / jnp.linalg.norm(quaternion_params)
    rotation_matrix = quaternions_to_rotation_matrix_jax(quaternion_params)

    # Construct SE(3) transformation matrix
    se3_matrix = jnp.eye(4)
    se3_matrix = se3_matrix.at[:3, :3].set(rotation_matrix)
    se3_matrix = se3_matrix.at[:3, 3].set(translation_params)
    return se3_matrix


def apply_SE3_transform_jax(points: Array,
                           SE3_transform: Array
                           ) -> Array:
    """
    Takes a point cloud and transforms it according to the provided SE3 transformation matrix.
    Jax implementation.

    Parameters
    ----------
    points : Array (N, 3)
        Set of coordinates representing a point cloud.
    SE3_transform : Array (4, 4)
        SE(3) transformation matrix.

    Returns
    -------
    transformed_points : Array (N, 3)
        Set of coordinates transformed by the corresponding SE(3) transformation.
    """
    # Single instance
    transformed_points = (SE3_transform[:3,:3] @ points.T).T + SE3_transform[:3,3]
    return transformed_points



"""
Autodock Vina Docking evaluation pipelines.

Requires:
- vina
- meeko
- openbabel (if protonating ligands)
"""
from typing import List, Optional, Dict, Literal, Tuple, Any
from pathlib import Path
from tqdm import tqdm
import multiprocessing
import pickle

import numpy as np
import pandas as pd
from rdkit import Chem
from rdkit.Chem import AllChem

from shepherd_score.evaluations.utils.convert_data import get_smiles_from_atom_pos

from shepherd_score.evaluations.docking.docking import VinaSmiles
from shepherd_score.evaluations.docking.targets import docking_target_info


# Global variable to store VinaSmiles instance in each worker process
_worker_vina_smiles = None


def _init_worker_vina(
    receptor_pdbqt_file: str,
    center: Tuple[float, float, float],
    box_size: Tuple[float, float, float],
    pH: float,
    scorefunction: str,
    num_processes: int,
    verbose: int,
):
    """
    Initialize VinaSmiles instance in worker process.
    This is called once per worker when the pool is created.
    """
    global _worker_vina_smiles
    _worker_vina_smiles = VinaSmiles(
        receptor_pdbqt_file=receptor_pdbqt_file,
        center=center,
        box_size=box_size,
        pH=pH,
        scorefunction=scorefunction,
        num_processes=num_processes,
        verbose=verbose,
    )


def _eval_docking_single(
    i: int,
    smiles: str,
    exhaustiveness: int,
    n_poses: int,
    protonate: bool,
    return_best_protomer: bool,
    save_poses_path: Optional[str],
) -> Dict[str, Any]:
    """
    Evaluate a single SMILES string for docking.

    This function is designed to be called by multiprocessing workers.
    It uses the pre-initialized VinaSmiles instance from the worker.
    """
    global _worker_vina_smiles

    if smiles is None:
        return {'i': i, 'energy': np.nan, 'error': 'SMILES is None'}

    if _worker_vina_smiles is None:
        return {'i': i, 'energy': np.nan, 'error': 'VinaSmiles not initialized in worker'}

    try:
        # Reset state before each docking call to ensure clean state
        # (set_ligand_from_string should replace previous ligand, but this ensures clean state)
        _worker_vina_smiles.state = None

        energy, docked_mol = _worker_vina_smiles(
            ligand_smiles=smiles,
            output_file=save_poses_path,
            exhaustiveness=exhaustiveness,
            n_poses=n_poses,
            protonate=protonate,
            return_best_protomer=True,
        )

        # Pickle the docked mol for return
        docked_mol_pickle = None
        if docked_mol is not None:
            docked_mol_pickle = pickle.dumps(docked_mol)

        return {
            'i': i,
            'energy': float(energy),
            'error': None,
            'docked_mol': docked_mol_pickle,
        }
    except Exception as e:
        return {'i': i, 'energy': np.nan, 'error': str(e), 'docked_mol': None}


def _unpack_eval_docking_single(args):
    """Unpacker function for multiprocessing."""
    return _eval_docking_single(*args)


def _eval_relax_single(
    i: int,
    mol_pickle: bytes,
    center: bool | Tuple[float, float, float],
    max_steps: int | None,
    save_poses_path: Optional[str],
) -> Dict[str, Any]:
    """
    Evaluate a single mol object for relaxation.

    This function is designed to be called by multiprocessing workers.
    It uses the pre-initialized VinaSmiles instance from the worker.
    """
    global _worker_vina_smiles

    if mol_pickle is None:
        return {'i': i, 'energy': np.nan, 'relaxed_mol': None, 'error': 'mol is None'}

    if _worker_vina_smiles is None:
        return {'i': i, 'energy': np.nan, 'relaxed_mol': None, 'error': 'VinaSmiles not initialized in worker'}

    try:
        # Unpickle the mol object
        mol = pickle.loads(mol_pickle)

        if mol is None:
            return {'i': i, 'energy': np.nan, 'relaxed_mol': None, 'error': 'mol is None after unpickling'}

        # Reset state before each optimization call
        _worker_vina_smiles.state = None

        total_energy, _, optimized_mol = _worker_vina_smiles.optimize_ligand(
            ligand=mol,
            center=center,
            max_steps=max_steps,
            output_file=save_poses_path,
        )

        # Pickle the optimized mol for return
        optimized_mol_pickle = None
        if optimized_mol is not None:
            optimized_mol_pickle = pickle.dumps(optimized_mol)

        return {
            'i': i,
            'energy': float(total_energy),
            'relaxed_mol': optimized_mol_pickle,
            'error': None
        }
    except Exception as e:
        return {'i': i, 'energy': np.nan, 'relaxed_mol': None, 'error': str(e)}


def _unpack_eval_relax_single(args):
    """Unpacker function for multiprocessing."""
    return _eval_relax_single(*args)


class DockingEvalPipeline:

    def __init__(self,
                 pdb_id: str,
                 num_processes: int = 4,
                 docking_target_info_dict: Dict = docking_target_info,
                 verbose: int = 0,
                 path_to_bin: str = ''):
        """
        Constructor for docking evaluation pipeline.

        Initializes VinaSmiles with receptor pdbqt.

        Parameters
        ----------
        pdb_id : str
            PDB ID of receptor. Natively only supports:
            1iep, 3eml, 3ny8, 4rlu, 4unn, 5mo4, 7l11.
        num_processes : int, optional
            Number of CPUs to use for scoring. Default is 4.
        docking_target_info_dict : dict, optional
            Dict holding minimum information needed for docking. Example format::

                {"1iep": {"center": (15.614, 53.380, 15.455),
                          "size": (15, 15, 15),
                          "pdbqt": "path_to_file.pdbqt"}}

        verbose : int, optional
            Level of verbosity from vina.Vina (0 is silent). Default is 0.
        path_to_bin : str, optional
            Path to environment bin containing ``mk_prepare_ligand.py``. Default is ''.
        """
        self.pdb_id = pdb_id
        self.path_to_bin = path_to_bin
        self.docking_target_info = docking_target_info_dict
        self.vina_smiles = None
        self.smiles = []
        self.energies = []
        self.buffer = {}
        self.num_failed = 0
        self.repeats = 0
        self.buffer_relaxed = {}
        self.relaxed_mols = []
        self.relaxed_rmsd = []

        if pdb_id not in list(self.docking_target_info.keys()):
            raise ValueError(
                f"Provided `pdb_id` ({pdb_id}) not supported. Please choose from: {list(self.docking_target_info.keys())}."
            )

        path_to_receptor_pdbqt = Path(self.docking_target_info[self.pdb_id]['pdbqt'])
        if not path_to_receptor_pdbqt.is_file():
            raise ValueError(
                f"Provided .pdbqt file does not exist. Please check `docking_target_info_dict`. Was given: {path_to_receptor_pdbqt}"
            )

        pH = self.docking_target_info[self.pdb_id]['pH'] if 'pH' in self.docking_target_info[self.pdb_id] else 7.4

        self.receptor_pdbqt_file = str(path_to_receptor_pdbqt)
        self.center = self.docking_target_info[self.pdb_id]['center']
        self.box_size = self.docking_target_info[self.pdb_id]['size']
        self.pH = pH
        self.scorefunction = 'vina'
        self.verbose = verbose

        self.vina_smiles = VinaSmiles(
            receptor_pdbqt_file=self.receptor_pdbqt_file,
            center=self.center,
            box_size=self.box_size,
            pH=self.pH,
            scorefunction=self.scorefunction,
            num_processes=num_processes,
            verbose=self.verbose
        )

    def evaluate(self,
                 smiles_ls: List[str],
                 exhaustiveness: int = 32,
                 n_poses: int = 1,
                 protonate: bool = False,
                 save_poses_dir_path: Optional[str] = None,
                 verbose: bool = False,
                 num_workers: int = 1,
                 num_processes: int = 4,
                 return_best_protomer: bool = False,
                 *,
                 mp_context: Literal['spawn', 'forkserver'] = 'spawn'
                 ) -> List[float]:
        """
        Loop through supplied list of SMILES strings, dock, and collect energies.

        Arguments
        ---------
        smiles_ls : List[str] list of SMILES to dock
        exhaustiveness : int (default = 32) Number of Monte Carlo simulations to run per pose
        n_poses : int (default = 1) Number of poses to save
        protonate : bool (default = False) Use protonation protocol
        save_poses_dir_path : Optional[str] (default = None) Path to directory to save docked poses.
        verbose : bool (default = False) show tqdm progress bar for each SMILES.
        num_workers : int (default = 1) number of parallel worker processes.
            Only recommended if `smiles_ls` is > 100 due to start-up overhead of new processes.
        num_processes : int (default = 4) number of processes each worker uses internally for Vina.
            Constraint: num_workers * num_processes <= available CPUs
        mp_context : Literal['spawn', 'forkserver'] context for multiprocessing.

        Returns
        -------
        List of energies (affinities) in kcal/mol
        """
        self.smiles = smiles_ls
        dir_path = None
        if save_poses_dir_path is not None:
            dir_path = Path(save_poses_dir_path)

        # Check buffer first and filter out cached SMILES
        energies = []
        indices_to_process = []
        smiles_to_process = []

        for i, smiles in enumerate(smiles_ls):
            if smiles in self.buffer:
                self.repeats += 1
                energies.append((i, self.buffer[smiles]['energy']))
            elif smiles is None:
                energies.append((i, np.nan))
                self.num_failed += 1
            else:
                indices_to_process.append(i)
                smiles_to_process.append(smiles)

        available_cpus = multiprocessing.cpu_count() or 1
        if num_processes > exhaustiveness:
            num_processes = exhaustiveness
        if num_workers < 1:
            num_workers = 1
        if num_processes < 1:
            num_processes = 1

        # Calculate max workers: num_workers * num_processes <= available_cpus
        max_workers_allowed = max(1, available_cpus // max(1, num_processes))
        if num_workers > max_workers_allowed:
            num_workers = max_workers_allowed

        if num_workers > 1 and len(smiles_to_process) > 0:
            multiprocessing.set_start_method(mp_context, force=True)

            # Prepare inputs for workers (only task-specific parameters)
            inputs = []
            for idx, smiles in zip(indices_to_process, smiles_to_process):
                save_poses_path = None
                if save_poses_dir_path is not None:
                    save_poses_path = str(dir_path / f'{self.pdb_id}_docked_{idx}{"_prot" if protonate else ""}.pdbqt')

                inputs.append((
                    idx,
                    smiles,
                    exhaustiveness,
                    n_poses,
                    protonate if not return_best_protomer else True,
                    return_best_protomer,
                    save_poses_path,
                ))

            # Initialize VinaSmiles once per worker using initializer
            with multiprocessing.Pool(
                num_workers,
                initializer=_init_worker_vina,
                initargs=(
                    self.receptor_pdbqt_file,
                    self.center,
                    self.box_size,
                    self.pH,
                    self.scorefunction,
                    num_processes,
                    self.verbose,
                )
            ) as pool:
                if verbose:
                    pbar = tqdm(total=len(smiles_to_process), desc=f'Docking {self.pdb_id}')
                else:
                    pbar = None

                results_iter = pool.imap_unordered(_unpack_eval_docking_single, inputs, chunksize=1)

                pending_results = {idx: None for idx in indices_to_process}

                for res in results_iter:
                    idx = res['i']
                    pending_results[idx] = res
                    energy = res['energy']
                    energies.append((idx, energy))

                    # Update buffer
                    smiles_str = smiles_ls[idx]
                    docked_mol = None
                    if res['docked_mol'] is not None:
                        try:
                            docked_mol = pickle.loads(res['docked_mol'])
                        except Exception:
                            docked_mol = None
                    if smiles_str is not None:
                        self.buffer[smiles_str] = {'energy': float(energy), 'docked_mol': docked_mol}
                        if return_best_protomer:
                            self.buffer[smiles_str].update({'protomer_smiles': None})
                            if docked_mol is not None:
                                self.buffer[smiles_str]['protomer_smiles'] = Chem.MolToSmiles(Chem.RemoveHs(docked_mol))
                        if np.isnan(energy):
                            self.num_failed += 1

                    if verbose and pbar is not None:
                        pbar.update(1)

                if verbose and pbar is not None:
                    pbar.close()

                for idx in indices_to_process:
                    if pending_results[idx] is None:
                        energies.append((idx, np.nan))
                        smiles_str = smiles_ls[idx]
                        if smiles_str is not None:
                            self.buffer[smiles_str] = {'energy': float(np.nan), 'docked_mol': None}
                            if return_best_protomer:
                                self.buffer[smiles_str].update({'protomer_smiles': None})
                        self.num_failed += 1
        else:
            # Single process evaluation
            if len(smiles_to_process) > 0:
                if verbose:
                    pbar = tqdm(enumerate(zip(indices_to_process, smiles_to_process)),
                            desc=f'Docking {self.pdb_id}',
                            total=len(smiles_to_process))
                else:
                    pbar = enumerate(zip(indices_to_process, smiles_to_process))

                for _, (idx, smiles) in pbar:
                    save_poses_path = None
                    if save_poses_dir_path is not None:
                        if return_best_protomer:
                            save_poses_path = dir_path / f'{self.pdb_id}_docked_best_prot_{idx}.pdbqt'
                        else:
                            save_poses_path = dir_path / f'{self.pdb_id}_docked{"_prot" if protonate else ""}_{idx}.pdbqt'
                    try:
                        energy, docked_mol = self.vina_smiles(
                            ligand_smiles=smiles,
                            output_file=save_poses_path,
                            exhaustiveness=exhaustiveness,
                            n_poses=n_poses,
                            protonate=protonate if not return_best_protomer else True,
                            return_best_protomer=return_best_protomer,
                        )
                        energies.append((idx, float(energy)))
                        self.buffer[smiles] = {'energy': float(energy), 'docked_mol': docked_mol}
                        if return_best_protomer:
                            self.buffer[smiles].update({'protomer_smiles': None})
                            if docked_mol is not None:
                                self.buffer[smiles]['protomer_smiles'] = Chem.MolToSmiles(Chem.RemoveHs(docked_mol))

                    except Exception:
                        energies.append((idx, np.nan))
                        self.buffer[smiles] = {'energy': float(np.nan), 'docked_mol': None}
                        if return_best_protomer:
                            self.buffer[smiles].update({'protomer_smiles': None})
                        self.num_failed += 1

        # Sort by original index and extract energies
        energies.sort(key=lambda x: x[0])
        self.energies = np.array([e[1] for e in energies])
        return [e[1] for e in energies]

    def evaluate_relax(self,
                       mol_ls: List[Chem.Mol],
                       center: bool | Tuple[float, float, float] = False,
                       max_steps: int | None = 10000,
                       save_poses_dir_path: Optional[str] = None,
                       verbose: bool = False,
                       num_workers: int = 1,
                       *,
                       mp_context: Literal['spawn', 'forkserver'] = 'spawn'
                       ) -> List[float]:
        """
        Loop through supplied list of mol objects, optimize, and collect energies.

        Arguments
        ---------
        mol_ls : List[Chem.Mol] list of rdkit mol objects to relax
        center : bool | Tuple[float, float, float] (default = False)
            If a tuple, centers to those coordinates.
            If True, centers the ligand to the receptor's center.
            If False, does not translate the ligand from its initial conformation.
        max_steps : int | None (default = 10000) Maximum number of steps to take in the optimization.
            If None, uses the default value of 10000.
        save_poses_dir_path : Optional[str] (default = None) Path to directory to save optimized poses.
        verbose : bool (default = False) show tqdm progress bar for each mol.
        num_workers : int (default = 1) number of parallel worker processes.
        mp_context : Literal['spawn', 'forkserver'] context for multiprocessing.

        Returns
        -------
        List of energies (affinities) in kcal/mol
        """
        dir_path = None
        if save_poses_dir_path is not None:
            dir_path = Path(save_poses_dir_path)

        # Process mol objects
        energies = []
        relaxed_mols = []
        indices_to_process = []
        mols_to_process = []

        for i, mol in enumerate(mol_ls):
            if mol is None:
                energies.append((i, np.nan))
                relaxed_mols.append((i, None))
                self.num_failed += 1
            else:
                indices_to_process.append(i)
                mols_to_process.append(mol)
            self.buffer_relaxed[Chem.MolToSmiles(Chem.RemoveHs(mol))] = {'energy': np.nan, 'relaxed_mol': None, 'rmsd': np.nan}

        available_cpus = multiprocessing.cpu_count() or 1
        if num_workers < 1:
            num_workers = 1

        # Calculate max workers: num_workers * num_processes <= available_cpus
        max_workers_allowed = max(1, available_cpus)
        if num_workers > max_workers_allowed:
            num_workers = max_workers_allowed

        if num_workers > 1 and len(mols_to_process) > 0:
            multiprocessing.set_start_method(mp_context, force=True)

            # Prepare inputs for workers (pickle mol objects for multiprocessing)
            inputs = []
            for idx, mol in zip(indices_to_process, mols_to_process):
                save_poses_path = None
                if save_poses_dir_path is not None:
                    save_poses_path = str(dir_path / f'{self.pdb_id}_relaxed_{idx}.pdbqt')

                # Pickle the mol for multiprocessing
                mol_pickle = pickle.dumps(mol)

                inputs.append((
                    idx,
                    mol_pickle,
                    center,
                    max_steps,
                    save_poses_path,
                ))

            # Initialize VinaSmiles once per worker using initializer
            with multiprocessing.Pool(
                num_workers,
                initializer=_init_worker_vina,
                initargs=(
                    self.receptor_pdbqt_file,
                    self.center,
                    self.box_size,
                    self.pH,
                    self.scorefunction,
                    1, # num processes = 1 since only one process is used for relaxation
                    self.verbose,
                )
            ) as pool:
                if verbose:
                    pbar = tqdm(total=len(mols_to_process), desc=f'Relaxing {self.pdb_id}')
                else:
                    pbar = None

                results_iter = pool.imap_unordered(_unpack_eval_relax_single, inputs, chunksize=1)

                pending_results = {idx: None for idx in indices_to_process}

                for res in results_iter:
                    idx = res['i']
                    pending_results[idx] = res
                    energy = res['energy']
                    energies.append((idx, energy))
                    smiles_str = Chem.MolToSmiles(Chem.RemoveHs(mols_to_process[idx]))
                    if smiles_str is not None:
                        self.buffer_relaxed[smiles_str]['energy'] = float(energy)
                        self.buffer_relaxed[smiles_str]['relaxed_mol'] = mols_to_process[idx]

                    # Unpickle the relaxed mol
                    relaxed_mol = None
                    if res['relaxed_mol'] is not None:
                        try:
                            relaxed_mol = pickle.loads(res['relaxed_mol'])
                        except Exception:
                            relaxed_mol = None
                    relaxed_mols.append((idx, relaxed_mol))
                    smiles_str = Chem.MolToSmiles(Chem.RemoveHs(mols_to_process[idx]))
                    if smiles_str is not None:
                        self.buffer_relaxed[smiles_str]['energy'] = float(energy)
                        self.buffer_relaxed[smiles_str]['relaxed_mol'] = relaxed_mol if relaxed_mol is not None else None
                        self.buffer_relaxed[smiles_str]['rmsd'] = AllChem.CalcRMS(
                            self.vina_smiles._center_ligand(mols_to_process[idx], self.center),
                            relaxed_mol
                        )
                    if np.isnan(energy):
                        self.num_failed += 1

                    if verbose and pbar is not None:
                        pbar.update(1)

                if verbose and pbar is not None:
                    pbar.close()

                for idx in indices_to_process:
                    if pending_results[idx] is None:
                        energies.append((idx, np.nan))
                        relaxed_mols.append((idx, None))
                        self.num_failed += 1
        else:
            # Single process evaluation
            if len(mols_to_process) > 0:
                if verbose:
                    pbar = tqdm(enumerate(zip(indices_to_process, mols_to_process)),
                            desc=f'Relaxing {self.pdb_id}',
                            total=len(mols_to_process))
                else:
                    pbar = enumerate(zip(indices_to_process, mols_to_process))

                for _, (idx, mol) in pbar:
                    save_poses_path = None
                    if save_poses_dir_path is not None:
                        save_poses_path = dir_path / f'{self.pdb_id}_relaxed_{idx}.pdbqt'
                    try:
                        total_energy, _, optimized_mol = self.vina_smiles.optimize_ligand(
                            ligand=mol,
                            center=center,
                            max_steps=max_steps,
                            output_file=save_poses_path,
                        )
                        energies.append((idx, float(total_energy)))
                        relaxed_mols.append((idx, optimized_mol))
                    except Exception:
                        energies.append((idx, np.nan))
                        relaxed_mols.append((idx, None))
                        self.num_failed += 1

                    smiles_str = Chem.MolToSmiles(Chem.RemoveHs(mol))
                    if smiles_str is not None:
                        self.buffer_relaxed[smiles_str]['energy'] = float(total_energy)
                        self.buffer_relaxed[smiles_str]['relaxed_mol'] = optimized_mol if optimized_mol is not None else None
                        self.buffer_relaxed[smiles_str]['rmsd'] = AllChem.CalcRMS(
                            self.vina_smiles._center_ligand(mol, self.center),
                            optimized_mol
                        )

        # Sort by original index and extract energies and relaxed mols
        energies.sort(key=lambda x: x[0])
        relaxed_mols.sort(key=lambda x: x[0])
        return np.array([e[1] for e in energies])

    def benchmark(self,
                  exhaustiveness: int = 32,
                  n_poses: int = 5,
                  protonate: bool = False,
                  save_poses_dir_path: Optional[str] = None
                  ) -> float:
        """
        Run benchmark with experimental ligands.

        Arguments
        ---------
        exhaustiveness : int (default = 32) Number of Monte Carlo simulations to run per pose
        n_poses : int (default = 5) Number of poses to save
        protonate : bool (default = False) (de-)protonate ligand with OpenBabel at pH=7.4
        save_poses_dir_path : Optional[str] (default = None) Path to directory to save docked poses.

        Returns
        -------
        float : Energies (affinities) in kcal/mol
        """
        save_poses_path = None
        if save_poses_dir_path is not None:
            dir_path = Path(save_poses_dir_path)
            save_poses_path = dir_path / f"{self.pdb_id}_docked{'_prot' if protonate else ''}.pdbqt"

        best_energy, docked_ligand = self.vina_smiles(
            self.docking_target_info[self.pdb_id]['ligand'],
            output_file=str(save_poses_path),
            exhaustiveness=exhaustiveness,
            n_poses=n_poses,
            protonate=protonate,
        )
        return best_energy, docked_ligand

    def to_pandas(self) -> pd.DataFrame:
        """
        Convert the attributes of generated smiles and the energies to a pd.DataFrame

        Returns
        -------
        pd.DataFrame : attributes for each evaluated sample
        """
        global_attrs = {'smiles' : self.smiles, 'energies': self.energies}
        series_global = pd.Series(global_attrs)

        return series_global

    def to_pandas_relaxed(self) -> pd.DataFrame:
        """
        Convert the attributes of relaxed mols and the energies to a pd.DataFrame

        Returns
        -------
        pd.DataFrame : attributes for each relaxed sample
        """
        df_relaxed = pd.DataFrame(self.buffer_relaxed)
        return df_relaxed


def run_docking_benchmark(save_dir_path: str,
                          pdb_id: str,
                          num_processes: int = 4,
                          docking_target_info_dict=docking_target_info,
                          protonate: bool = False
                          ) -> None:
    """
    Run docking benchmark on experimental SMILES.

    Uses an exhaustiveness of 32 and saves the top-30 poses to a specified location.

    Parameters
    ----------
    save_dir_path : str
        Path to save docked poses to.
    pdb_id : str
        PDB ID of receptor. Natively only supports:
        1iep, 3eml, 3ny8, 4rlu, 4unn, 5mo4, 7l11.
    num_processes : int, optional
        Number of CPUs to use for scoring. Default is 4.
    docking_target_info_dict : dict, optional
        Dict holding minimum information needed for docking. Example format::

            {"1iep": {"center": (15.614, 53.380, 15.455),
                      "size": (15, 15, 15),
                      "pdbqt": "path_to_file.pdbqt",
                      "ligand": "SMILES string of experimental ligand"}}

    protonate : bool, optional
        Whether to protonate ligands at a given pH. Requires ``"pH"`` field to be
        filled out in docking_target_info_dict. Default is ``False``.

    Returns
    -------
    None
    """
    dep = DockingEvalPipeline(pdb_id=pdb_id,
                              num_processes=num_processes,
                              docking_target_info_dict=docking_target_info_dict,
                              verbose=0,
                              path_to_bin='')
    dep.benchmark(exhaustiveness=32, n_poses=30, save_poses_dir_path=save_dir_path, protonate=protonate)


def run_docking_evaluation(atoms: List[np.ndarray],
                           positions: List[np.ndarray],
                           pdb_id: str,
                           num_processes: int = 4,
                           docking_target_info_dict=docking_target_info,
                           exhaustiveness: int = 32,
                           n_poses: int = 1,
                           protonate: bool = False,
                           save_poses_dir_path: Optional[str] = None,
                           verbose: bool = True,
                           num_workers: int = 1,
                           *,
                           mp_context: Literal['spawn', 'forkserver'] = 'spawn'
                           ) -> DockingEvalPipeline:
    """
    Run docking evaluation with an exhaustiveness of 32.

    Parameters
    ----------
    atoms : list
        List of np.ndarray (N,) of atomic numbers of the generated molecule or (N, M)
        one-hot encoding.
    positions : list
        List of np.ndarray (N, 3) of coordinates for the generated molecule's atoms.
    pdb_id : str
        PDB ID of receptor. Natively only supports:
        1iep, 3eml, 3ny8, 4rlu, 4unn, 5mo4, 7l11.
    num_processes : int, optional
        Number of CPUs to use for Autodock Vina. Default is 4.
    docking_target_info_dict : dict, optional
        Dict holding minimum information needed for docking. Example format::

            {"1iep": {"center": (15.614, 53.380, 15.455),
                      "size": (15, 15, 15),
                      "pdbqt": "path_to_file.pdbqt"}}

    exhaustiveness : int, optional
        Number of Monte Carlo simulations to run per pose. Default is 32.
    n_poses : int, optional
        Number of poses to save. Default is 1.
    protonate : bool, optional
        Use protonation protocol. Default is ``False``.
    save_poses_dir_path : str, optional
        Path to directory to save docked poses. Default is ``None``.
    verbose : bool, optional
        Show tqdm progress bar for each SMILES. Default is ``True``.
    num_workers : int, optional
        Number of parallel worker processes. Default is 1.
    mp_context : str, optional
        Context for multiprocessing. One of 'spawn' or 'forkserver'. Default is 'spawn'.

    Returns
    -------
    DockingEvalPipeline
        Results are found in the ``buffer`` attribute {'smiles': energy} or in ``smiles``
        and ``energies`` which preserves the order of provided atoms/positions as a list.
    """
    docking_pipe = DockingEvalPipeline(pdb_id=pdb_id,
                                       num_processes=num_processes,
                                       docking_target_info_dict=docking_target_info_dict,
                                       verbose=0,
                                       path_to_bin='')

    smiles_list = []
    for sample in zip(atoms, positions):
        smiles_list.append(get_smiles_from_atom_pos(atoms=sample[0], positions=sample[1]))

    docking_pipe.evaluate(smiles_list, exhaustiveness=exhaustiveness, n_poses=n_poses,
                          protonate=protonate, save_poses_dir_path=save_poses_dir_path,
                          verbose=verbose, num_workers=num_workers, num_processes=num_processes,
                          mp_context=mp_context)

    return docking_pipe


def run_docking_evaluation_from_smiles(smiles: List[str],
                                       pdb_id: str,
                                       num_processes: int = 4,
                                       docking_target_info_dict=docking_target_info,
                                       exhaustiveness: int = 32,
                                       n_poses: int = 1,
                                       protonate: bool = False,
                                       save_poses_dir_path: Optional[str] = None,
                                       verbose: bool = True,
                                       num_workers: int = 1,
                                       *,
                                       mp_context: Literal['spawn', 'forkserver'] = 'spawn'
                                       ) -> DockingEvalPipeline:
    """
    Run docking evaluation with an exhaustiveness of 32.

    Parameters
    ----------
    smiles : list
        List of SMILES strings. These must be valid or ``None``.
    pdb_id : str
        PDB ID of receptor. Natively only supports:
        1iep, 3eml, 3ny8, 4rlu, 4unn, 5mo4, 7l11.
    num_processes : int, optional
        Number of CPUs to use for Autodock Vina. Default is 4.
    docking_target_info_dict : dict, optional
        Dict holding minimum information needed for docking. Example format::

            {"1iep": {"center": (15.614, 53.380, 15.455),
                      "size": (15, 15, 15),
                      "pdbqt": "path_to_file.pdbqt"}}

    exhaustiveness : int, optional
        Number of Monte Carlo simulations to run per pose. Default is 32.
    n_poses : int, optional
        Number of poses to save. Default is 1.
    protonate : bool, optional
        Use protonation protocol. Default is ``False``.
    save_poses_dir_path : str, optional
        Path to directory to save docked poses. Default is ``None``.
    verbose : bool, optional
        Show tqdm progress bar for each SMILES. Default is ``True``.
    num_workers : int, optional
        Number of parallel worker processes. Default is 1.
    mp_context : str, optional
        Context for multiprocessing. One of 'spawn' or 'forkserver'. Default is 'spawn'.

    Returns
    -------
    DockingEvalPipeline
        Results are found in the ``buffer`` attribute {'smiles': energy} or in ``smiles``
        and ``energies`` which preserves the order of provided SMILES as a list.
    """
    docking_pipe = DockingEvalPipeline(pdb_id=pdb_id,
                                       num_processes=num_processes,
                                       docking_target_info_dict=docking_target_info_dict,
                                       verbose=0,
                                       path_to_bin='')

    docking_pipe.evaluate(smiles, exhaustiveness=exhaustiveness, n_poses=n_poses,
                          protonate=protonate, save_poses_dir_path=save_poses_dir_path,
                          verbose=verbose, num_workers=num_workers, num_processes=num_processes,
                          mp_context=mp_context)

    return docking_pipe



"""
Torch implementations for principal component alignment (pca). Written to handle batching.

IT IS RECOMMENDED TO USE THE NUMPY VERSION.
Using the numpy version of quaternions_for_principal_component_alignment is faster
(~2.5ms vs ~5ms). Further, the parallel GPU version is slightly slower (50us) than the serial,
CPU version.

Credit to Lewis J. Martin as this was adapted from
https://github.com/ljmartin/align/blob/main/0.2%20aligning%20principal%20moments%20of%20inertia.ipynb
"""
import torch
import torch.nn.functional as F
from shepherd_score.alignment_utils.se3 import get_SE3_transform, apply_SE3_transform

def compute_moment_of_inertia(points: torch.Tensor) -> torch.Tensor:
    """
    Computes the moment of inertia of a set of points.
    A = x^2 + y^2 + z^2
    B = X^T X
    """
    # Single instance
    if len(points.shape) == 2:
        # Translate points to center of mass
        translated_points = points - torch.mean(points, dim=0)

        A = torch.sum(translated_points**2)
        B = translated_points.T @ translated_points
        return (A * torch.eye(3).to(points.device) - B) / points.shape[0]

    # Batched
    elif len(points.shape) == 3:
        batch_size = points.shape[0]
        # center to COM
        translated_points = points - torch.mean(points, dim=1).unsqueeze(1)

        A = torch.sum(translated_points**2, dim=1).sum(dim=1)
        B = torch.bmm(translated_points.permute(0,2,1), translated_points)
        A_eye = (A.unsqueeze(1)
                 * torch.eye(3).flatten().repeat((batch_size,1)).to(points.device)
                ).reshape((batch_size, 3, 3))
        return (A_eye - B) / points.shape[1]

    else:
        raise ValueError(f'Expected "points" to have shape (batch, N, 3), or (N, 3), but {points.shape} was passed.')


def compute_principal_moments_of_interia(points: torch.Tensor) -> torch.Tensor:
    """ Compute the principal moments of inertia of a set of points. """
    moment_of_inertia_tensor = compute_moment_of_inertia(points)
    # Eigvals are sorted in ascending order
    _, eigvecs = torch.linalg.eigh(moment_of_inertia_tensor)
    # Single Instance
    if len(points.shape) == 2:
        return torch.flip(eigvecs, (1,)).T
    # Batched
    elif len(points.shape) == 3:
        return torch.flip(eigvecs, (2,)).permute(0,2,1)
    else:
        raise ValueError(f'Expected "points" to have shape (batch, N, 3), or (N, 3), but {points.shape} was passed.')


def angle_between_vecs(v1, v2):
    """ Compute the angle in radians between two vectors (already normalized). """
    # Single Instance
    if len(v1.shape) == 1 and len(v2.shape) == 1:
        return torch.acos(torch.clamp(torch.dot(v1, v2), min=-1., max=1.)) # radians
    # Batched
    elif len(v1.shape) == 2 and len(v2.shape) == 2:
        return torch.acos(torch.clamp(torch.sum(v1 * v2, dim=1), min=-1., max=1.)).unsqueeze(1)
    else:
        raise ValueError(f'Expected "v1" and "v2" to have shape (batch, 3), or (3,), but {v1.shape} and {v2.shape} was passed.')


def rotation_axis(v1, v2):
    """
    Calculate the vector about which to order to rotate `a` to align with `b` (cross product).
    """
    # Single Instance
    if len(v1.shape) == 1 and len(v2.shape) == 1:
        if torch.allclose(v1, v2):
            return torch.Tensor([1, 0, 0]).to(v1.device)
        v3 = torch.linalg.cross(v1, v2, dim=0)

    # Batched
    elif len(v1.shape) == 2 and len(v2.shape) == 2:
        idx_not_same = torch.where(torch.isclose(v1, v2).sum(1) != 3)[0]
        v3 = torch.zeros((v1.shape[0], 3)).to(v1.device)
        v3[:, 0] = 1.
        v3[idx_not_same] = torch.linalg.cross(v1, v2, dim=1)
    else:
        raise ValueError(f'Expected "v1" and "v2" to have shape (batch, 3), or (3,), but {v1.shape} and {v2.shape} was passed.')

    return F.normalize(v3, p=2, dim=len(v1.shape)-1)


def quaternion_from_axis_angle(axis, angle):
    """
    Create a Quaternion from a rotation axis and an angle in radians.

    Parameters
    ----------
    axis : torch.Tensor (3,)
        Axis to rotate about.
    angle: torch.Tensor (1,)
        Angle in radians.

    Returns
    -------
    quaternion : torch.Tensor (4,)
    """
    # Single Instance
    if len(axis.shape) == 1:
        mag_sq = torch.dot(axis, axis)
        if not torch.is_nonzero(mag_sq):
            raise ZeroDivisionError("Provided rotation axis has no length")
        theta = angle / 2.0
        r = torch.cos(theta)
        i = axis * torch.sin(theta)
        return torch.Tensor([r, i[0], i[1], i[2]]).to(axis.device)

    # Batched
    elif len(axis.shape) == 2 and len(angle) == axis.shape[0]:
        mag_sq = torch.sum(axis ** 2, dim=1, keepdim=True)
        if torch.any(mag_sq == 0):
            raise ZeroDivisionError("Provided rotation axis has no length")
        theta = angle / 2.0
        r = torch.cos(theta)
        i = axis * torch.sin(theta)
        return torch.cat((r, i), dim=1).to(axis.device)
    else:
        raise ValueError(f'Expected "axis" and "angle" to have corresponding shapes (batch, 3)+(batch,1), or (3,)+(1,), but {axis.shape} and {angle.shape} was passed.')


def quaternion_mult(p: torch.Tensor, q: torch.Tensor) -> torch.Tensor:
    """
    Multiplication of quaternions p and q.

    Reference: https://academicflight.com/articles/kinematics/rotation-formalisms/quaternions/

    General use case: The consecutive rotations of q_1 then q_2 is equivalent
    to q_3 = q_2*q_1. (order matters)

    Parameters
    ----------
    p : torch.Tensor
        The first quaternion with shape (4,) or (batch, 4).
    q : torch.Tensor
        The second quaternion with shape (4,) or (batch, 4).

    Returns
    -------
    torch.Tensor
        The product of the two quaternions with shape (4,) or (batch, 4).
    """
    if len(p.shape) == 1 and len(q.shape) == 1:
        mat1 = torch.Tensor([[p[0], -p[1], -p[2], -p[3]],
                             [p[1],  p[0], -p[3],  p[2]],
                             [p[2],  p[3],  p[0], -p[1]],
                             [p[3], -p[2],  p[1],  p[0]]]).to(p.device)
        pq = mat1 @ q
    elif len(p.shape) == 2 and len(q.shape)==2:
        pq = torch.empty_like(p).to(p.device)
        pq[:, 0] = p[:, 0] * q[:, 0] - p[:, 1] * q[:, 1] - p[:, 2] * q[:, 2] - p[:, 3] * q[:, 3]
        pq[:, 1] = p[:, 0] * q[:, 1] + p[:, 1] * q[:, 0] + p[:, 2] * q[:, 3] - p[:, 3] * q[:, 2]
        pq[:, 2] = p[:, 0] * q[:, 2] - p[:, 1] * q[:, 3] + p[:, 2] * q[:, 0] + p[:, 3] * q[:, 1]
        pq[:, 3] = p[:, 0] * q[:, 3] + p[:, 1] * q[:, 2] - p[:, 2] * q[:, 1] + p[:, 3] * q[:, 0]
    else:
        raise ValueError(f'Expected "p" and "q" to have the same shape (batch, 4), or (4,), but {p.shape} and {q.shape} was passed.')
    return pq


def quaternions_for_principal_component_alignment(ref_points: torch.Tensor, fit_points: torch.Tensor) -> torch.Tensor:
    """
    Computes the 4 quaternions required for alignment of the fit mol along the
    principal components of the reference mol.

    The computed quaternions assumes that fit_points will be rotated after being centered at COM.
    """
    pmi_ref = compute_principal_moments_of_interia(ref_points)

    # If CPU compute with for-loops
    if ref_points.get_device() == -1:
        quaternions = torch.zeros((4,4)).to(ref_points.device)
        for q_index in range(4):
            if q_index == 1:
                # flip orientation of longest axis
                pmi_ref[0] = -pmi_ref[0]
            elif q_index == 2:
                # unflip orientation of longest axis
                pmi_ref[0] = -pmi_ref[0]
                # flip orientation of 2nd longest axis
                pmi_ref[1] = -pmi_ref[1]
            elif q_index == 3:
                # flip orientation of both axes
                pmi_ref[0] = -pmi_ref[0]

            quat_order = torch.zeros((2,4))
            # Initially center to COM
            fit_points_adjust = fit_points - torch.mean(fit_points, dim=0)
            for ax_idx in range(2):
                pmi_fit = compute_principal_moments_of_interia(fit_points_adjust)
                # Angle between principal axis of fit mol and referencne mol
                angle = angle_between_vecs(pmi_fit[ax_idx], pmi_ref[ax_idx])
                # Axis that we are rotating about
                ax = rotation_axis(pmi_fit[ax_idx], pmi_ref[ax_idx])
                # Quaternion
                quat_order[ax_idx] = quaternion_from_axis_angle(ax, angle)
                # get SE(3) transformation matrix
                se3_params = torch.concatenate((quat_order[ax_idx], torch.zeros(3)))
                # get transformed matrix
                fit_points_adjust = apply_SE3_transform(fit_points_adjust, get_SE3_transform(se3_params))
            quaternions[q_index] = quaternion_mult(quat_order[1], quat_order[0])

    else: # GPU
        pmi_refs = pmi_ref.repeat((4, 1, 1))
        pmi_refs[1][0] = -pmi_refs[1][0] # flip orientation of longest axis
        pmi_refs[2][1] = -pmi_refs[2][1] # flip orientation of 2nd longest axis
        # flip orientation of longest and 2nd longest axes
        pmi_refs[3][0] = -pmi_refs[3][0]
        pmi_refs[3][1] = -pmi_refs[3][1]

        fit_points_adjust = fit_points.repeat((4,1,1))
        # Initially center to COM
        fit_points_adjust = fit_points_adjust - torch.mean(fit_points_adjust, dim=1).unsqueeze(1)
        quat_order = torch.zeros((8,4))
        for ax_idx in range(2):
            # Principal moment of inertia of molecule getting aligned
            pmi_fit = compute_principal_moments_of_interia(fit_points_adjust)
            # Angle between principal axis of fit mol and referencne mol
            angle = angle_between_vecs(pmi_fit[:, ax_idx], pmi_refs[:, ax_idx])
            # Axis that we are rotating about
            ax = rotation_axis(pmi_fit[:, ax_idx], pmi_refs[:, ax_idx])
            # Quaternion
            quat_order[ax_idx*4:(ax_idx+1)*4] = quaternion_from_axis_angle(ax, angle)
            # get SE(3) transformation matrix
            se3_params = torch.concatenate((quat_order[ax_idx*4:(ax_idx+1)*4], torch.zeros((4,3))), axis=1).to(ref_points.device)
            # get transformed matrix
            fit_points_adjust = apply_SE3_transform(fit_points_adjust, get_SE3_transform(se3_params))
        quaternions = quaternion_mult(quat_order[4:], quat_order[:4])
    return quaternions



"""
Cross-platform test suite for pharmacophore scoring implementations.
Tests PyTorch and JAX implementations against NumPy ground truth.
"""
import pytest
import numpy as np
import warnings
from typing import Tuple

# Always import NumPy implementation (ground truth)
from shepherd_score.score.pharmacophore_scoring_np import (
    get_overlap_pharm_np,
    get_pharm_combo_score as get_pharm_combo_score_np
)

from .utils import _configure_jax_platform

# Pre-configure JAX platform
_gpu_detected_early = _configure_jax_platform()

# Try to import all implementations
try:
    import torch
    TORCH_AVAILABLE = True
except ImportError:
    TORCH_AVAILABLE = False
    warnings.warn("PyTorch not available - skipping PyTorch tests")

try:
    import jax
    import jax.numpy as jnp
    JAX_AVAILABLE = True

    # Check for GPU more safely now
    try:
        gpu_devices = jax.devices('gpu')
        GPU_AVAILABLE = len(gpu_devices) > 0 and _gpu_detected_early
    except Exception:
        GPU_AVAILABLE = False

except ImportError:
    JAX_AVAILABLE = False
    GPU_AVAILABLE = False
    warnings.warn("JAX not available - skipping JAX tests")

# Import implementations conditionally
if TORCH_AVAILABLE:
    from shepherd_score.score.pharmacophore_scoring import (
        get_overlap_pharm as get_overlap_pharm_torch,
        get_pharm_combo_score as get_pharm_combo_score_torch
    )

if JAX_AVAILABLE:
    from shepherd_score.score.pharmacophore_scoring_jax import (
        get_overlap_pharm_jax,
        get_overlap_pharm_jax_vectorized,
        get_pharm_combo_score_jax
    )

# Test tolerances
RTOL = 1e-5
ATOL = 1e-7

class TestDataGenerator:
    """Generate consistent test data across all implementations."""

    @staticmethod
    def generate_pharmacophore_data(n_pharm_1: int = 10, n_pharm_2: int = 8,
                                  seed: int = 42) -> Tuple[np.ndarray, ...]:
        """Generate test pharmacophore data."""
        np.random.seed(seed)

        # Pharmacophore types (indices 0-7 for the 8 pharmacophore types)
        ptype_1 = np.random.randint(0, 8, size=n_pharm_1)
        ptype_2 = np.random.randint(0, 8, size=n_pharm_2)

        # Anchor coordinates
        anchors_1 = np.random.normal(0, 5, size=(n_pharm_1, 3)).astype(np.float32)
        anchors_2 = np.random.normal(0, 5, size=(n_pharm_2, 3)).astype(np.float32)

        # Vector directions (normalized)
        vectors_1 = np.random.normal(0, 1, size=(n_pharm_1, 3)).astype(np.float32)
        vectors_1 = vectors_1 / np.linalg.norm(vectors_1, axis=1, keepdims=True)

        vectors_2 = np.random.normal(0, 1, size=(n_pharm_2, 3)).astype(np.float32)
        vectors_2 = vectors_2 / np.linalg.norm(vectors_2, axis=1, keepdims=True)

        return ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2

    @staticmethod
    def generate_shape_data(n_atoms_1: int = 20, n_atoms_2: int = 15,
                          seed: int = 42) -> Tuple[np.ndarray, np.ndarray]:
        """Generate test shape data (atomic coordinates)."""
        np.random.seed(seed + 100)  # Different seed for shape data

        centers_1 = np.random.normal(0, 3, size=(n_atoms_1, 3)).astype(np.float32)
        centers_2 = np.random.normal(0, 3, size=(n_atoms_2, 3)).astype(np.float32)

        return centers_1, centers_2

# Fixtures for test data
@pytest.fixture(params=[
    (5, 3),    # Small
    (10, 8),   # Medium
    (20, 15),  # Large
    (1, 1),    # Single pharmacophore each
])
def pharmacophore_sizes(request):
    """Different pharmacophore array sizes for testing."""
    return request.param

@pytest.fixture(params=[
    'tanimoto',
    'tversky',
    'tversky_ref',
    'tversky_fit'
])
def similarity_func(request):
    """Different similarity functions for testing."""
    return request.param

@pytest.fixture(params=[
    (False, False),  # No extended points
    (True, False),   # Extended points, include anchors
    (True, True),    # Extended points only
])
def extended_points_config(request):
    """Different extended points configurations."""
    extended_points, only_extended = request.param
    return extended_points, only_extended

@pytest.fixture
def test_data(pharmacophore_sizes):
    """Generate test data for given sizes."""
    n1, n2 = pharmacophore_sizes
    data = TestDataGenerator.generate_pharmacophore_data(n1, n2)
    shape_data = TestDataGenerator.generate_shape_data(n1 + 5, n2 + 5)
    return data + shape_data

class TestPharmacophoreScoring:
    """Test pharmacophore scoring implementations."""

    def test_single_pharmacophore_type(self):
        """Test with all pharmacophores of the same type."""
        # All hydrophobic pharmacophores
        ptype_1 = np.array([0, 0, 0], dtype=np.int32)
        ptype_2 = np.array([0, 0], dtype=np.int32)

        anchors_1 = np.random.normal(0, 2, size=(3, 3)).astype(np.float32)
        anchors_2 = np.random.normal(0, 2, size=(2, 3)).astype(np.float32)
        vectors_1 = np.random.normal(0, 1, size=(3, 3)).astype(np.float32)
        vectors_2 = np.random.normal(0, 1, size=(2, 3)).astype(np.float32)

        # Normalize vectors
        vectors_1 = vectors_1 / np.linalg.norm(vectors_1, axis=1, keepdims=True)
        vectors_2 = vectors_2 / np.linalg.norm(vectors_2, axis=1, keepdims=True)

        # NumPy result
        result_np = get_overlap_pharm_np(
            ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2
        )

        # Compare with other implementations
        if TORCH_AVAILABLE:
            result_torch = get_overlap_pharm_torch(
                torch.from_numpy(ptype_1), torch.from_numpy(ptype_2),
                torch.from_numpy(anchors_1), torch.from_numpy(anchors_2),
                torch.from_numpy(vectors_1), torch.from_numpy(vectors_2)
            )
            assert np.allclose(result_torch.numpy(), result_np, rtol=RTOL, atol=ATOL)

        if JAX_AVAILABLE:
            result_jax = get_overlap_pharm_jax(
                jnp.array(ptype_1), jnp.array(ptype_2),
                jnp.array(anchors_1), jnp.array(anchors_2),
                jnp.array(vectors_1), jnp.array(vectors_2)
            )
            assert np.allclose(np.array(result_jax), result_np, rtol=RTOL, atol=ATOL)

    @pytest.mark.parametrize("similarity", ['tanimoto', 'tversky', 'tversky_ref', 'tversky_fit'])
    def test_pharmacophore_scoring_similarity_functions(self, test_data, similarity):
        """Test different similarity functions."""
        ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2, centers_1, centers_2 = test_data

        # NumPy result (ground truth)
        result_np = get_overlap_pharm_np(
            ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2,
            similarity=similarity
        )

        # PyTorch comparison
        if TORCH_AVAILABLE:
            result_torch = get_overlap_pharm_torch(
                torch.from_numpy(ptype_1), torch.from_numpy(ptype_2),
                torch.from_numpy(anchors_1), torch.from_numpy(anchors_2),
                torch.from_numpy(vectors_1), torch.from_numpy(vectors_2),
                similarity=similarity
            )
            assert np.allclose(result_torch.numpy(), result_np, rtol=RTOL, atol=ATOL), \
                f"PyTorch result {result_torch.numpy()} != NumPy result {result_np} for {similarity}"

        # JAX comparison
        if JAX_AVAILABLE:
            result_jax = get_overlap_pharm_jax(
                jnp.array(ptype_1), jnp.array(ptype_2),
                jnp.array(anchors_1), jnp.array(anchors_2),
                jnp.array(vectors_1), jnp.array(vectors_2),
                similarity=similarity
            )
            assert np.allclose(np.array(result_jax), result_np, rtol=RTOL, atol=ATOL), \
                f"JAX result {np.array(result_jax)} != NumPy result {result_np} for {similarity}"

    def test_extended_points_configurations(self, test_data, extended_points_config):
        """Test extended points configurations."""
        extended_points, only_extended = extended_points_config
        ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2, centers_1, centers_2 = test_data

        # NumPy result (ground truth)
        result_np = get_overlap_pharm_np(
            ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2,
            extended_points=extended_points, only_extended=only_extended
        )

        # PyTorch comparison
        if TORCH_AVAILABLE:
            result_torch = get_overlap_pharm_torch(
                torch.from_numpy(ptype_1), torch.from_numpy(ptype_2),
                torch.from_numpy(anchors_1), torch.from_numpy(anchors_2),
                torch.from_numpy(vectors_1), torch.from_numpy(vectors_2),
                extended_points=extended_points, only_extended=only_extended
            )
            assert np.allclose(result_torch.numpy(), result_np, rtol=RTOL, atol=ATOL), \
                f"PyTorch result != NumPy result for extended_points={extended_points}, only_extended={only_extended}"

        # JAX comparison
        if JAX_AVAILABLE:
            result_jax = get_overlap_pharm_jax(
                jnp.array(ptype_1), jnp.array(ptype_2),
                jnp.array(anchors_1), jnp.array(anchors_2),
                jnp.array(vectors_1), jnp.array(vectors_2),
                extended_points=extended_points, only_extended=only_extended
            )
            assert np.allclose(np.array(result_jax), result_np, rtol=RTOL, atol=ATOL), \
                f"JAX result != NumPy result for extended_points={extended_points}, only_extended={only_extended}"

class TestComboScoring:
    """Test combined pharmacophore + shape scoring."""

    def test_combo_scoring_basic(self, test_data):
        """Test basic combo scoring functionality."""
        ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2, centers_1, centers_2 = test_data

        # NumPy result (ground truth)
        result_np = get_pharm_combo_score_np(
            centers_1, centers_2, ptype_1, ptype_2,
            anchors_1, anchors_2, vectors_1, vectors_2
        )

        # PyTorch comparison
        if TORCH_AVAILABLE:
            result_torch = get_pharm_combo_score_torch(
                torch.from_numpy(centers_1), torch.from_numpy(centers_2),
                torch.from_numpy(ptype_1), torch.from_numpy(ptype_2),
                torch.from_numpy(anchors_1), torch.from_numpy(anchors_2),
                torch.from_numpy(vectors_1), torch.from_numpy(vectors_2)
            )
            assert np.allclose(result_torch.numpy(), result_np, rtol=RTOL, atol=ATOL)

        # JAX comparison
        if JAX_AVAILABLE:
            result_jax = get_pharm_combo_score_jax(
                jnp.array(centers_1), jnp.array(centers_2),
                jnp.array(ptype_1), jnp.array(ptype_2),
                jnp.array(anchors_1), jnp.array(anchors_2),
                jnp.array(vectors_1), jnp.array(vectors_2)
            )
            assert np.allclose(np.array(result_jax), result_np, rtol=RTOL, atol=ATOL)

class TestEdgeCases:
    """Test edge cases and error conditions."""

    def test_mismatched_pharmacophore_types(self):
        """Test with completely different pharmacophore types."""
        # Molecule 1 has only hydrophobic (0)
        ptype_1 = np.array([0, 0, 0], dtype=np.int32)
        # Molecule 2 has only aromatic (4)
        ptype_2 = np.array([4, 4], dtype=np.int32)

        anchors_1 = np.random.normal(0, 1, size=(3, 3)).astype(np.float32)
        anchors_2 = np.random.normal(0, 1, size=(2, 3)).astype(np.float32)
        vectors_1 = np.random.normal(0, 1, size=(3, 3)).astype(np.float32)
        vectors_2 = np.random.normal(0, 1, size=(2, 3)).astype(np.float32)

        # Normalize vectors
        vectors_1 = vectors_1 / np.linalg.norm(vectors_1, axis=1, keepdims=True)
        vectors_2 = vectors_2 / np.linalg.norm(vectors_2, axis=1, keepdims=True)

        # Should give low/zero similarity
        result_np = get_overlap_pharm_np(
            ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2
        )

        # Result should be very small (likely 0) since no overlapping pharmacophore types
        assert result_np >= 0.0  # Similarity should be non-negative

        # Test consistency across implementations
        if TORCH_AVAILABLE:
            result_torch = get_overlap_pharm_torch(
                torch.from_numpy(ptype_1), torch.from_numpy(ptype_2),
                torch.from_numpy(anchors_1), torch.from_numpy(anchors_2),
                torch.from_numpy(vectors_1), torch.from_numpy(vectors_2)
            )
            assert np.allclose(result_torch.numpy(), result_np, rtol=RTOL, atol=ATOL)

        if JAX_AVAILABLE:
            result_jax = get_overlap_pharm_jax(
                jnp.array(ptype_1), jnp.array(ptype_2),
                jnp.array(anchors_1), jnp.array(anchors_2),
                jnp.array(vectors_1), jnp.array(vectors_2)
            )
            assert np.allclose(np.array(result_jax), result_np, rtol=RTOL, atol=ATOL)

    def test_identical_molecules(self):
        """Test scoring identical molecules (should give similarity = 1.0)."""
        ptype_1 = np.array([0, 1, 2, 4], dtype=np.int32)  # Mix of types
        anchors_1 = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float32)
        vectors_1 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0]], dtype=np.float32)
        vectors_1 = vectors_1 / np.linalg.norm(vectors_1, axis=1, keepdims=True)

        # Use same data for both molecules
        ptype_2, anchors_2, vectors_2 = ptype_1, anchors_1, vectors_1

        result_np = get_overlap_pharm_np(
            ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2,
            similarity='tanimoto'
        )

        # Should be exactly 1.0 for identical molecules with Tanimoto similarity
        assert np.allclose(result_np, 1.0, rtol=1e-10, atol=1e-10)

        # Test other implementations
        if TORCH_AVAILABLE:
            result_torch = get_overlap_pharm_torch(
                torch.from_numpy(ptype_1), torch.from_numpy(ptype_2),
                torch.from_numpy(anchors_1), torch.from_numpy(anchors_2),
                torch.from_numpy(vectors_1), torch.from_numpy(vectors_2),
                similarity='tanimoto'
            )
            assert np.allclose(result_torch.numpy(), 1.0, rtol=1e-10, atol=1e-10)

        if JAX_AVAILABLE:
            result_jax = get_overlap_pharm_jax(
                jnp.array(ptype_1), jnp.array(ptype_2),
                jnp.array(anchors_1), jnp.array(anchors_2),
                jnp.array(vectors_1), jnp.array(vectors_2),
                similarity='tanimoto'
            )
            assert np.allclose(np.array(result_jax), 1.0, rtol=1e-10, atol=1e-10)

@pytest.mark.skipif(not JAX_AVAILABLE, reason="JAX not available")
class TestVectorizedPharmScoring:
    """
    Verify that get_overlap_pharm_jax_vectorized produces results numerically
    equivalent to get_overlap_pharm_jax across all similarity modes, extended-
    point configurations, and pharmacophore-size combinations.

    The vectorized implementation uses a single (N, M) broadcast instead of
    per-type loops, so any discrepancy would indicate an indexing or mode-
    mapping bug rather than a numerical precision issue.
    """

    # Looser than the scoring-only tolerance because the vectorized scorer
    # re-normalises vectors internally (divides by norm + 1e-6).
    ATOL = 1e-4
    RTOL = 1e-4

    @pytest.mark.parametrize("similarity", ['tanimoto', 'tversky_ref', 'tversky_fit'])
    @pytest.mark.parametrize("sizes", [(5, 3), (10, 8), (1, 1)])
    def test_matches_reference_scorer(self, similarity, sizes):
        """Vectorized scorer matches the reference JAX scorer for different
        similarity functions and pharmacophore counts."""
        n1, n2 = sizes
        ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2 = \
            TestDataGenerator.generate_pharmacophore_data(n1, n2, seed=7)

        p1 = jnp.array(ptype_1)
        p2 = jnp.array(ptype_2)
        a1 = jnp.array(anchors_1)
        a2 = jnp.array(anchors_2)
        v1 = jnp.array(vectors_1)
        v2 = jnp.array(vectors_2)

        ref = float(get_overlap_pharm_jax(p1, p2, a1, a2, v1, v2, similarity=similarity))
        vec = float(get_overlap_pharm_jax_vectorized(p1, p2, a1, a2, v1, v2))

        # get_overlap_pharm_jax_vectorized only returns the raw VAB overlap;
        # similarity is handled externally in the optimisation loop.  Here we
        # compare VAB directly, so use get_overlap_pharm_jax with tanimoto and
        # back out VAB by checking self-scores.
        vaa = float(get_overlap_pharm_jax_vectorized(p1, p1, a1, a1, v1, v1))
        vbb = float(get_overlap_pharm_jax_vectorized(p2, p2, a2, a2, v2, v2))
        eps = 1e-6
        if similarity == 'tanimoto':
            vec_score = vec / (vaa + vbb - vec + eps)
        elif similarity == 'tversky_ref':
            vec_score = vec / (vaa + eps)
        else:  # tversky_fit
            vec_score = vec / (vbb + eps)

        assert np.isclose(vec_score, ref, atol=self.ATOL, rtol=self.RTOL), (
            f"Vectorized score {vec_score:.6f} != reference {ref:.6f} "
            f"(similarity={similarity}, sizes={sizes})"
        )

    @pytest.mark.parametrize("extended_points,only_extended", [
        (False, False),
        (True,  False),
        (True,  True),
    ])
    def test_extended_points_matches_reference(self, extended_points, only_extended):
        """Vectorized scorer matches the reference scorer for extended-point modes."""
        ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2 = \
            TestDataGenerator.generate_pharmacophore_data(8, 6, seed=13)

        p1 = jnp.array(ptype_1)
        p2 = jnp.array(ptype_2)
        a1 = jnp.array(anchors_1)
        a2 = jnp.array(anchors_2)
        v1 = jnp.array(vectors_1)
        v2 = jnp.array(vectors_2)

        ref = float(get_overlap_pharm_jax(
            p1, p2, a1, a2, v1, v2,
            similarity='tanimoto',
            extended_points=extended_points,
            only_extended=only_extended,
        ))

        vab = float(get_overlap_pharm_jax_vectorized(
            p1, p2, a1, a2, v1, v2,
            extended_points=extended_points, only_extended=only_extended,
        ))
        vaa = float(get_overlap_pharm_jax_vectorized(
            p1, p1, a1, a1, v1, v1,
            extended_points=extended_points, only_extended=only_extended,
        ))
        vbb = float(get_overlap_pharm_jax_vectorized(
            p2, p2, a2, a2, v2, v2,
            extended_points=extended_points, only_extended=only_extended,
        ))
        eps = 1e-6
        vec_score = vab / (vaa + vbb - vab + eps)

        assert np.isclose(vec_score, ref, atol=self.ATOL, rtol=self.RTOL), (
            f"Vectorized score {vec_score:.6f} != reference {ref:.6f} "
            f"(extended_points={extended_points}, only_extended={only_extended})"
        )

    def test_identical_molecules_score_one(self):
        """Self-overlap normalised to Tanimoto similarity must equal 1.0."""
        ptype_1, _, anchors_1, _, vectors_1, _ = \
            TestDataGenerator.generate_pharmacophore_data(6, 4, seed=99)

        p = jnp.array(ptype_1)
        a = jnp.array(anchors_1)
        v = jnp.array(vectors_1)

        vab = float(get_overlap_pharm_jax_vectorized(p, p, a, a, v, v))
        vaa = float(get_overlap_pharm_jax_vectorized(p, p, a, a, v, v))
        # vaa == vbb == vab  →  Tanimoto = vab / (2*vaa - vab) = 1.0
        eps = 1e-6
        tanimoto = vab / (vaa + vaa - vab + eps)
        assert np.isclose(tanimoto, 1.0, atol=1e-5), (
            f"Self-similarity {tanimoto:.8f} != 1.0"
        )

    def test_jit_cache_stability(self):
        """Calling the vectorized scorer repeatedly with identical shapes must
        not grow the JIT cache (no recompilation after the first call)."""
        import jax
        ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2 = \
            TestDataGenerator.generate_pharmacophore_data(8, 6, seed=5)

        p1 = jnp.array(ptype_1)
        p2 = jnp.array(ptype_2)
        a1 = jnp.array(anchors_1)
        a2 = jnp.array(anchors_2)
        v1 = jnp.array(vectors_1)
        v2 = jnp.array(vectors_2)

        # Warm-up
        _ = get_overlap_pharm_jax_vectorized(p1, p2, a1, a2, v1, v2)

        compilations_before = sum(
            getattr(fn, '_cache_size', lambda: 0)()
            for fn in [get_overlap_pharm_jax_vectorized]
        )

        for _ in range(5):
            _ = get_overlap_pharm_jax_vectorized(p1, p2, a1, a2, v1, v2)

        compilations_after = sum(
            getattr(fn, '_cache_size', lambda: 0)()
            for fn in [get_overlap_pharm_jax_vectorized]
        )

        assert compilations_after == compilations_before, (
            "JIT cache grew during repeated calls with identical shapes — "
            "unexpected recompilation detected."
        )


# Performance benchmarking (optional, for development)
class TestPerformance:
    """Performance comparison tests (marked as slow)."""

    @pytest.mark.slow
    @pytest.mark.skipif(not (TORCH_AVAILABLE and JAX_AVAILABLE),
                       reason="Both PyTorch and JAX needed for performance comparison")
    def test_performance_comparison(self):
        """Compare performance of different implementations."""
        import time

        # Generate larger test data
        ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2 = \
            TestDataGenerator.generate_pharmacophore_data(50, 40, seed=42)

        n_runs = 10

        # NumPy timing
        np_times = []
        for _ in range(n_runs):
            start = time.time()
            result_np = get_overlap_pharm_np(
                ptype_1, ptype_2, anchors_1, anchors_2, vectors_1, vectors_2
            )
            np_times.append(time.time() - start)
        np_time = np.mean(np_times)

        # PyTorch timing
        torch_ptype_1 = torch.from_numpy(ptype_1)
        torch_ptype_2 = torch.from_numpy(ptype_2)
        torch_anchors_1 = torch.from_numpy(anchors_1)
        torch_anchors_2 = torch.from_numpy(anchors_2)
        torch_vectors_1 = torch.from_numpy(vectors_1)
        torch_vectors_2 = torch.from_numpy(vectors_2)

        torch_times = []
        for _ in range(n_runs):
            start = time.time()
            result_torch = get_overlap_pharm_torch(
                torch_ptype_1, torch_ptype_2, torch_anchors_1, torch_anchors_2,
                torch_vectors_1, torch_vectors_2
            )
            torch_times.append(time.time() - start)
        torch_time = np.mean(torch_times)

        # JAX timing (with warmup)
        jax_ptype_1 = jnp.array(ptype_1)
        jax_ptype_2 = jnp.array(ptype_2)
        jax_anchors_1 = jnp.array(anchors_1)
        jax_anchors_2 = jnp.array(anchors_2)
        jax_vectors_1 = jnp.array(vectors_1)
        jax_vectors_2 = jnp.array(vectors_2)

        # Warmup for JIT compilation
        get_overlap_pharm_jax(
            jax_ptype_1, jax_ptype_2, jax_anchors_1, jax_anchors_2,
            jax_vectors_1, jax_vectors_2
        ).block_until_ready()

        jax_times = []
        for _ in range(n_runs):
            start = time.time()
            result_jax = get_overlap_pharm_jax(
                jax_ptype_1, jax_ptype_2, jax_anchors_1, jax_anchors_2,
                jax_vectors_1, jax_vectors_2
            ).block_until_ready()
            jax_times.append(time.time() - start)
        jax_time = np.mean(jax_times)

        print(f"\nPerformance comparison ({n_runs} runs):")
        print(f"NumPy:   {np_time:.4f}s")
        print(f"PyTorch: {torch_time:.4f}s ({torch_time/np_time:.2f}x vs NumPy)")
        print(f"JAX:     {jax_time:.4f}s ({jax_time/np_time:.2f}x vs NumPy)")

        # Verify results are still consistent
        assert np.allclose(result_torch.numpy(), result_np, rtol=RTOL, atol=ATOL)
        assert np.allclose(np.array(result_jax), result_np, rtol=RTOL, atol=ATOL)

if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
Functions used for SE(3) transformations. (NumPy implementation).

Namely, converting quaternions to rotation matrices, getting an SE(3) transform from SE(3)
parameters, and applying the SE(3) transformation on a set of points.

Credit to Lewis J. Martin as this was adapted from
https://github.com/ljmartin/align/blob/main/0.2%20aligning%20principal%20moments%20of%20inertia.ipynb
and PyTorch's implementations.
"""
import numpy as np

def quaternions_to_rotation_matrix_np(quaternions: np.ndarray) -> np.ndarray:
    """
    Converts quaternion to a rotation matrix.
    Adapted from PyTorch3D:
    https://pytorch3d.readthedocs.io/en/latest/_modules/pytorch3d/transforms/rotation_conversions.html#quaternion_to_matrix

    Parameters
    ----------
    quaternions : np.ndarray (4,)
        Quaternion parameters in (r,i,j,k) order.
        set.

    Returns
    -------
    rotation_matrix : np.ndarray (3,3)
        Rotation matrix converted from quaternion.
    """
    # Single instance
    if quaternions.shape == (4,):
        r, i, j, k = quaternions
        two_s = 2. / (quaternions * quaternions).sum()

        rotation_matrix = np.stack(
            (
                1 - two_s * (j * j + k * k),
                two_s * (i * j - k * r),
                two_s * (i * k + j * r),
                two_s * (i * j + k * r),
                1 - two_s * (i * i + k * k),
                two_s * (j * k - i * r),
                two_s * (i * k - j * r),
                two_s * (j * k + i * r),
                1 - two_s * (i * i + j * j),
            ),
            -1,
        )
    else:
        raise ValueError(f'Input "quaternions" must be a 1D Tensor of length 4. Instead the shape given was: {quaternions.shape}')
    return rotation_matrix.reshape((3, 3))


def get_SE3_transform_np(se3_params: np.ndarray
                         ) -> np.ndarray:
    """
    Constructs an SE(3) transformtion matrix from parameters. NumPy implementation

    Parameters
    ----------
    se3_params : np.ndarray (7,)
        Parameters for SE(3) transformation.
        The first 4 values in the last dimension are quaternions of form (r,i,j,k)
        and the last 3 values of the last dimension are the translations in (x,y,z).

    Returns
    -------
    se3_matrix : np.ndarray (4, 4)
        se3_params converted to a 4x4 SE(3) transformation matrix.
    """
    if se3_params.shape == (7,):
        # Extract quaternion and translation parameters
        quaternion_params = se3_params[:4]
        translation_params = se3_params[4:]

        # Normalize quaternion to ensure unit length
        quaternion_params = quaternion_params / np.linalg.norm(quaternion_params)
        rotation_matrix = quaternions_to_rotation_matrix_np(quaternion_params)

        # Construct SE(3) transformation matrix
        se3_matrix = np.eye(4)
        se3_matrix[:3, :3] = rotation_matrix
        se3_matrix[:3, 3] = translation_params
    else:
        raise ValueError(f'Input "se3_params" must be a 1D Tensor of length 7. Instead the shape given was: {se3_params.shape}')
    return se3_matrix


def apply_SE3_transform_np(points: np.ndarray,
                           SE3_transform: np.ndarray
                           ) -> np.ndarray:
    """
    Takes a point cloud and transforms it according to the provided SE3 transformation matrix.
    NumPy implementation.

    Parameters
    ----------
    points : np.ndarray (N, 3)
        Set of coordinates representing a point cloud.
    SE3_transform : np.ndarray (4, 4)
        SE(3) transformation matrix.

    Returns
    -------
    transformed_points : np.ndarray (N, 3)
        Set of coordinates transformed by the corresponding SE(3) transformation.
    """
    if points.shape[-1] != 3:
        raise ValueError(f'"points" should have shape (N_points, 3). Instead the shape given was: {points.shape}')
    if SE3_transform.shape[-2:] != (4,4):
        raise ValueError(f'"SE3_transform" should have shape (4, 4). Instead the shape given was: {SE3_transform.shape}')
    if len(SE3_transform.shape) != len(points.shape):
        raise ValueError(f'Shapes of points and SE3_transform should be the same length. Instead {len(SE3_transform.shape)} and {len(points.shape)} were given.')

    # Single instance
    if len(SE3_transform.shape) == 2:
        transformed_points = (SE3_transform[:3,:3] @ points.T).T + SE3_transform[:3,3]
    else:
        raise ValueError(f'The expected length of shape for "points" and "SE3_transform" must be 2 but {len(SE3_transform)} was given.')
    return transformed_points


def apply_SO3_transform_np(points: np.ndarray,
                           SE3_transform: np.ndarray
                           ) -> np.ndarray:
    """
    Takes a point cloud and ONLY ROTATES it according to the provided SE3 transformation matrix.
    Supports batched and non-batched inputs.

    Parameters
    ----------
    points : np.array (N, 3)
        Set of coordinates representing a point cloud.
    SE3_transform : (4, 4)
        SE(3) transformation matrix. If 'points' argument is batched, this one should be too.

    Returns
    -------
    rotated_points : torch.Tensor (batch, N, 3) or (N, 3)
        Set of coordinates rotated by the rotation component of the SE(3) transformation.
    """
    if points.shape[-1] != 3:
        raise ValueError(f'"points" should have shape (N_points, 3). Instead the shape given was: {points.shape}')
    if SE3_transform.shape[-2:] != (4,4):
        raise ValueError(f'"SE3_transform" should have shape (4, 4). Instead the shape given was: {SE3_transform.shape}')
    if len(SE3_transform.shape) != len(points.shape):
        raise ValueError(f'Shapes of points and SE3_transform should be the same length. Instead {len(SE3_transform.shape)} and {len(points.shape)} were given.')

    # Single instance
    if len(SE3_transform.shape) == 2:
        rotated_points = (SE3_transform[:3,:3] @ points.T).T
    else:
        raise ValueError(f'"points" and "SE3_transform" must be a single instance. \
        The expected length of shape for both should be 2 single instance but {len(SE3_transform)} was given.')

    return rotated_points



"""
Gaussian volume overlap scoring functions combined with continuous electrostatics
PYTORCH VERSION.

Batched and non-batched functionalities

Reference math:
https://doi.org/10.1002/(SICI)1096-987X(19961115)17:14<1653::AID-JCC7>3.0.CO;2-K
https://doi.org/10.1021/j100011a016
"""
import numpy as np
import torch
from shepherd_score.score.constants import COULOMB_SCALING, LAM_SCALING
from shepherd_score.score.gaussian_overlap import get_overlap


def VAB_2nd_order_esp(centers_1: torch.Tensor,
                      centers_2: torch.Tensor,
                      charges_1: torch.Tensor,
                      charges_2: torch.Tensor,
                      alpha: float,
                      lam: float
                      ) -> torch.Tensor:
    """ Torch implementation. Handles batching"""
    R2_sq = torch.cdist(centers_1, centers_2)**2.0
    C2_sq = torch.cdist(charges_1, charges_2)**2.0

    if R2_sq.dim() == 2: # Both centers_1 and centers_2 were single instances
        # R2_sq and C2_sq have shape (N, M)
        VAB_2nd_order = torch.sum(np.pi**(1.5) \
                                  / ((2*alpha)**(1.5)) \
                                  * torch.exp(-(alpha / 2) * R2_sq) \
                                  * torch.exp(-C2_sq/lam))

    elif R2_sq.dim() == 3: # At least one of centers_1 or centers_2 was batched
        # R2_sq and C2_sq have shape (B, N, M) or (N, B, M) if c1 was batched and c2 single (not typical for cdist)
        # More typically, if c1=(N,D) c2=(B,M,D) -> (B,N,M)
        # Or if c1=(B,N,D) c2=(M,D) -> (B,N,M)
        # Or if c1=(B,N,D) c2=(B,M,D) -> (B,N,M)
        # We permute to (B, M, N) to match the batched summation logic
        R2 = R2_sq.permute(0,2,1)
        C2 = C2_sq.permute(0,2,1)

        VAB_2nd_order = torch.sum(torch.sum(np.pi**(1.5) \
                                            / ((2*alpha)**(1.5)) \
                                            * torch.exp(-(alpha / 2) * R2) \
                                            * torch.exp(-C2/lam),
                                            dim = 2), # Sum over N
                                    dim = 1) # Sum over M
    else:
        raise ValueError(f"Unexpected dimension for R2_sq: {R2_sq.dim()}. centers_1: {centers_1.shape}, centers_2: {centers_2.shape}")

    return VAB_2nd_order

def shape_tanimoto_esp(centers_1: torch.Tensor,
                       centers_2: torch.Tensor,
                       charges_1: torch.Tensor,
                       charges_2: torch.Tensor,
                       alpha: float,
                       lam: float
                       ) -> torch.Tensor:
    """ Compute Tanimoto shape similarity """
    VAA = VAB_2nd_order_esp(centers_1, centers_1, charges_1, charges_1, alpha, lam)
    VBB = VAB_2nd_order_esp(centers_2, centers_2, charges_2, charges_2, alpha, lam)
    VAB = VAB_2nd_order_esp(centers_1, centers_2, charges_1, charges_2, alpha, lam)
    return VAB / (VAA + VBB - VAB)


def get_overlap_esp(centers_1: torch.Tensor,
                    centers_2: torch.Tensor,
                    charges_1: torch.Tensor,
                    charges_2: torch.Tensor,
                    alpha: float = 0.81,
                    lam: float = 0.3*LAM_SCALING
                    ) -> torch.Tensor:
    """
    Torch implementation.
    Compute electrostatic similarity which weights Gaussian volume overlap by electrostatics.
    The Tanimoto score is used.

    Typically `lam=0.3*LAM_SCALING` is used for surface point clouds and `lam=0.1` for partial charge
    weighted volumetric overlap.

    Parameters
    ----------
    centers_1 : torch.Tensor (batch, N, 3) or (N, 3)
        Coordinates for the sets of points representing molecule 1.
    centers_2 : torch.Tensor (batch, N, 3) or (N, 3)
        Coordinates for the sets of points representing molecule 2.
    charges_1 : torch.Tensor (batch, N) or (N,)
        Electrostatic energy for the sets of points representing molecule 1.
    charges_2 : torch.Tensor (batch, N) or (N,)
        Electrostatic energy for the sets of points representing molecule 2.
    alpha : float
        Parameter controlling the width of the Gaussians.
    lam : float
        Parameter controlling the influence of electrostatics.

    Returns
    -------
    tanimoto_esp : torch.Tensor (batch, 1) or (1,)
        Tanimoto similarities of electrostatics.
    """
    if isinstance(centers_1, np.ndarray):
        centers_1 = torch.Tensor(centers_1)
    if isinstance(centers_2, np.ndarray):
        centers_2 = torch.Tensor(centers_2)
    if isinstance(charges_1, np.ndarray):
        charges_1 = torch.Tensor(charges_1)
    if isinstance(charges_2, np.ndarray):
        charges_2 = torch.Tensor(charges_2)
    if len(charges_1.shape) == 1:
        charges_1 = charges_1.reshape((-1,1))
    elif len(charges_1.shape) == 2:
        charges_1 = charges_1.unsqueeze(2)
    if len(charges_2.shape) == 1:
        charges_2 = charges_2.reshape((-1,1))
    elif len(charges_2.shape) == 2:
        charges_2 = charges_2.unsqueeze(2)

    tanimoto_esp = shape_tanimoto_esp(centers_1, centers_2,
                                      charges_1, charges_2,
                                      alpha,
                                      lam)
    return tanimoto_esp


def _esp_comparison(points_1: torch.Tensor,
                    centers_w_H_2: torch.Tensor,
                    partial_charges_2: torch.Tensor,
                    points_charges_1: torch.Tensor,
                    radii_2: torch.Tensor,
                    probe_radius: float = 1.0,
                    lam: float = 0.001
                    ) -> torch.Tensor:
    """
    Helper function for computing the electrostatic potential (ESP) component of ShaEP score.
    It computes the difference in ESP at surface/observer points of molecule 1 for the ESP values
    generated by molecule 1 and molecule 2. It masks out observer points if they are in
    molecule 2's volume defined by vdW+probe_radius.
    Expects single instance or batched. This will ONLY check the shape of the points_1 to deterimine
    if it is batched or not so errors in the shape of the other tensors may or may not be caught.

    Parameters
    ----------
    points_1 : torch.Tensor (N_surf, 3) or (batch, N_surf, 3)
        Surface points of molecule 1 for which ESP's will be computed and compared.

    centers_w_H_2 : torch.Tensor (M + m_H, 3) or (batch, M + m_H, 3)
        Coordinates for atoms (including hydrogens) of molecule 2. Used in calculation of ESP at
        points_1 and masking out those within molecule 2's volume.

    partial_charges_2 : torch.Tensor (M + m_H,) or (batch, M + m_H,)
        Partial charges corresponding to centers_w_H_2. Used to calculate ESP.

    points_charges_1 : torch.Tensor (N_surf,) or (batch, N_surf,)
        Precalculated ESP's of molecule 1 corresponding to points_1.

    radii_2 : torch.Tensor (M + m_H,) or (batch, M + m_H,)
        Radii of each atom corresponding to centers_w_H_2. Used for masking operation.

    probe_radius : float (default = 1.0)
        Probe radius (default is 1 angstrom). Surfaces assumed to be generated with vdW radius and
        a probe radius of 1.2 angstroms (vdW radius of hydrogen). 1.0 used rather than 1.2 as a
        tolerance.

    lam : float (default = 0.001)
        Electrostatic potential weighting parameter (smaller = higher weight).
        0.001 was chosen as default based empirical observations of the distribution of scores
        generated before the summation in this function.

    Returns
    -------
    torch.Tensor (1,) or (batch, 1)
        Point to point ESP comparison. Scores range: [0, N_surf]. Score decreases for differences
        in ESP or due to masking of poorly aligned surface points.
    """
    lam = LAM_SCALING * lam
    distances = torch.cdist(points_1, centers_w_H_2)
    # single instance path for points_1
    if len(points_1.shape) == 2:
        if distances.dim() == 2: # centers_w_H_2 is also single (M+m_H, 3)
            mask = torch.where(torch.all(distances >= radii_2 + probe_radius, axis=1), 1., 0.) # (N_surf,)
            esp_at_surf_1 = torch.matmul(partial_charges_2, 1 / distances.T) * COULOMB_SCALING # (N_surf,)
            esp_diff_sq = torch.square(points_charges_1 - esp_at_surf_1) # (N_surf,)
            esp = torch.sum(mask * torch.exp(-esp_diff_sq/lam)) # scalar

        elif distances.dim() == 3: # centers_w_H_2 is batched (B, M+m_H, 3)

            # mask out molecule 1 surface points that are within molecule 2 (batched)
            mask = torch.where(torch.all(distances >= radii_2.unsqueeze(1) + probe_radius, axis=2), 1., 0.) # (B, N_surf)

            esp_at_surf_1 = torch.matmul(partial_charges_2.unsqueeze(1), 1 / distances.permute(0,2,1)) * COULOMB_SCALING # (B, 1, N_surf)

            esp_diff_sq = torch.square(points_charges_1.unsqueeze(0) - esp_at_surf_1.squeeze(1)) # (B, N_surf) after broadcasting and squeeze

            esp = torch.sum(mask * torch.exp(-esp_diff_sq/lam), axis=1) # (B,)
        else:
            raise ValueError(f"Distances tensor has unexpected dimensions {distances.dim()} when points_1 is single.")

    # batched path for points_1
    elif len(points_1.shape) == 3: # points_1 is (B, N_surf, 3)
        # This case assumes centers_w_H_2 and other inputs are compatibly batched or single (will broadcast)
        # distances will be (B, N_surf, M+m_H) irrespective of centers_w_H_2 being (B,M,3) or (M,3)
        _radii_2 = radii_2
        _partial_charges_2 = partial_charges_2

        if centers_w_H_2.shape[0] != points_1.shape[0] and centers_w_H_2.dim() == 3:
             raise ValueError(f"centers_w_H_2 has unexpected shape {centers_w_H_2.shape} when points_1 is batched. points_1: {points_1.shape}.")

        if _radii_2.dim() == 1:
            _radii_2 = _radii_2.unsqueeze(0)
        if _partial_charges_2.dim() == 1:
            _partial_charges_2 = _partial_charges_2.unsqueeze(0)

        # mask out molecule 1 surface points that are within molecule 2
        mask = torch.where(torch.all(distances >= _radii_2.unsqueeze(1) + probe_radius, axis=2), 1., 0.) # (B,N)

        # Calculate the potentials
        esp_at_surf_1 = torch.matmul(_partial_charges_2.unsqueeze(1), 1 / distances.permute(0,2,1)) * COULOMB_SCALING # (B,1,N)

        esp_diff_sq = torch.square(points_charges_1.unsqueeze(1) - esp_at_surf_1) # (B,1,N)
        esp = torch.sum(mask * torch.exp(-esp_diff_sq/lam).squeeze(1), axis=1) # (B,)
    else:
        raise ValueError(f"points_1 has unexpected shape {points_1.shape}")

    return esp


def esp_combo_score(centers_w_H_1: torch.Tensor,
                    centers_w_H_2: torch.Tensor,
                    centers_1: torch.Tensor,
                    centers_2: torch.Tensor,
                    points_1: torch.Tensor,
                    points_2: torch.Tensor,
                    partial_charges_1: torch.Tensor,
                    partial_charges_2: torch.Tensor,
                    point_charges_1: torch.Tensor,
                    point_charges_2: torch.Tensor,
                    radii_1: torch.Tensor,
                    radii_2: torch.Tensor,
                    alpha: float,
                    lam: float=0.001,
                    probe_radius: float=1.0,
                    esp_weight: float=0.5
                    ) -> torch.Tensor:
    """
    Computes a similarity score defined by ShaEP. It is a balanced score between electrostatics
    and shape similarity.
    Single instance or batch accepted (in the 0th dimension).
    This will ONLY check the shape of points_1 to deterimine if it is batched or not so errors
    in the shape of the other tensors may or may not be caught.

    Parameters
    ----------
    centers_w_H_1 : torch.Tensor (N + n_H, 3) | (batch, N + n_H, 3)
        Coordinates of atom centers INCLUDING hydrogens of molecule 1.
        Used for computing electrostatic potential.
        Same for centers_w_H_2 except (M + m_H, 3).

    centers_1 : torch.Tensor (N, 3) or (n_surf, 3) | (batch, N, 3) or (batch, n_surf, 3)
        Coordinates of points for molecule 1 used to compute shape similarity.
        Use atom centers for volumentric similarity. Use surface centers for surface similarity.
        Same for centers except (M, 3) or (m_surf, 3).

    points_1 : torch.Tensor (n_surf, 3) | (batch, n_surf, 3)
        Coordinates of surface points for molecule 1.
        Same for points_2 except (m_surf, 3).

    partial_charges_1 : torch.Tensor (N + n_H,) | (batch, N + n_H,)
        Partial charges corresponding to the atoms in centers_w_H_1.
        Same for partial_charges_2 except (M + m_H,).

    point_charges_1 : torch.Tensor (n_surf,) | (batch, n_surf,)
        The electrostatic potential calculated at each surface point (points_1).
        Same for point_charges_1 except (m_surf,)

    radii_1 : torch.Tensor (N + n_H,) | (batch, N + n_H,)
        vdW radii corresponding to the atoms in centers_w_H_1 (angstroms)
        Same for radii_2 except (M + m_H,)

    alpha : float
        Gaussian width parameter used for shape similarity.

    lam : float (default = 0.001)
        Electrostatic potential weighting parameter (smaller = higher weight).
        0.001 was chosen as default based empirical observations of the distribution of scores
        generated by _esp_comparison before summation.

    probe_radius : float (default = 1.0)
        Surface points found within vdW radii + probe radius will be masked out. Surface generation
        uses a probe radius of 1.2 (radius of hydrogen) so we use a slightly lower radius for be
        more tolerant.

    esp_weight : float (default = 0.5)
        Weight to be placed on electrostatic similarity with respect to shape similarity.
        0 = only shape similarity
        1 = only electrostatic similarity

    Returns
    -------
    torch.Tensor (1,) or (batch, 1)
        Similarity score (range: [0, 1]). Higher is more similar.
    """

    # Calculate the difference in ESP at the surface of molecule 1
    #   Expects hydrogens for the centers
    esp_1 = _esp_comparison(points_1, centers_w_H_2, partial_charges_2, point_charges_1, radii_2, probe_radius, lam)
    esp_2 = _esp_comparison(points_2, centers_w_H_1, partial_charges_1, point_charges_2, radii_1, probe_radius, lam)

    # Determine number of points for normalization, robust to single or batched inputs
    _num_points_1 = points_1.shape[1] if points_1.dim() == 3 else points_1.shape[0]
    _num_points_2 = points_2.shape[1] if points_2.dim() == 3 else points_2.shape[0]

    electrostatic_sim = (esp_1 + esp_2) / (_num_points_1 + _num_points_2)

    volumetric_sim = get_overlap(centers_1, centers_2, alpha)

    score = esp_weight*electrostatic_sim + (1-esp_weight)*volumetric_sim
    return score



"""
Module contains evaluation pipeline classes for generated molecules.
"""

from shepherd_score.evaluations.evaluate.evals import ConfEval, ConsistencyEval, ConditionalEval
from shepherd_score.evaluations.evaluate.pipelines import UnconditionalEvalPipeline
from shepherd_score.evaluations.evaluate.pipelines import ConditionalEvalPipeline, ConsistencyEvalPipeline
from shepherd_score.evaluations.evaluate.pipelines import resample_surf_scores, get_mol_from_atom_pos

__all__ = [
    'UnconditionalEvalPipeline',
    'ConditionalEvalPipeline',
    'ConsistencyEvalPipeline',
    'ConfEval',
    'ConsistencyEval',
    'ConditionalEval',
    'resample_surf_scores',
    'get_mol_from_atom_pos'
]



"""
Pharmacophore scoring with PyTorch.
"""
from typing import Union, Callable, Literal, Tuple
from functools import partial

import numpy as np
import torch
import torch.nn.functional as F

from shepherd_score.score.gaussian_overlap import VAB_2nd_order, VAB_2nd_order_mask, VAB_2nd_order_cosine
from shepherd_score.score.gaussian_overlap import VAB_2nd_order_mask_batch, VAB_2nd_order_cosine_mask_batch
from shepherd_score.score.constants import P_TYPES, P_ALPHAS
P_TYPES_LWRCASE = tuple(map(str.lower, P_TYPES))


def _compute_ref_overlap(overlap_func: Callable,
                         anchors_1: torch.Tensor,
                         alpha: float,
                         vectors_1: Union[torch.Tensor, None] = None,
                         allow_antiparallel: bool = False,
                         mask_1: Union[torch.Tensor, None] = None,
                         ) -> torch.Tensor:
    """ Both single instance and batched """
    # Just anchor volume overlap
    if (vectors_1 is None) and (mask_1 is None):
        VAA = overlap_func(anchors_1, anchors_1, alpha)
    # Just anchor volume overlap with masking for batching
    elif (vectors_1 is None) and (mask_1 is not None):
        VAA = overlap_func(anchors_1, anchors_1, alpha, mask_1=mask_1, mask_2=mask_1)
    # Anchor and vector volume overlap with masking for batching
    elif mask_1 is not None:
        VAA = overlap_func(anchors_1, anchors_1, vectors_1, vectors_1, alpha, allow_antiparallel, mask_1=mask_1, mask_2=mask_1)
    # Anchor and vector volume overlap for single instance
    else:
        VAA = overlap_func(anchors_1, anchors_1, vectors_1, vectors_1, alpha, allow_antiparallel)
    return VAA


def _compute_fit_overlap(overlap_func: Callable,
                         anchors_2: torch.Tensor,
                         alpha: float,
                         vectors_2: Union[torch.Tensor, None]  = None,
                         allow_antiparallel: bool = False,
                         mask_2: Union[torch.Tensor, None] = None
                         ) -> torch.Tensor:
    """ Both single instance and batched """
    # Just anchor volume overlap
    if (vectors_2 is None) and (mask_2 is None):
        VBB = overlap_func(anchors_2, anchors_2, alpha)
    # Just anchor volume overlap with masking for batching
    elif (vectors_2 is None) and (mask_2 is not None):
        VBB = overlap_func(anchors_2, anchors_2, alpha, mask_1=mask_2, mask_2=mask_2)
    # Anchor and vector volume overlap with masking for batching
    elif mask_2 is not None:
        VBB = overlap_func(anchors_2, anchors_2, vectors_2, vectors_2, alpha, allow_antiparallel, mask_1=mask_2, mask_2=mask_2)
    # Anchor and vector volume overlap for single instance
    else:
        VBB = overlap_func(anchors_2, anchors_2, vectors_2, vectors_2, alpha, allow_antiparallel)
    return VBB


def _compute_all_overlaps(overlap_func: Callable,
                          anchors_1: torch.Tensor,
                          anchors_2: torch.Tensor,
                          alpha: float,
                          vectors_1: Union[torch.Tensor, None] = None,
                          vectors_2: Union[torch.Tensor, None]  = None,
                          allow_antiparallel: bool = False,
                          mask_1: Union[torch.Tensor, None] = None,
                          mask_2: Union[torch.Tensor, None] = None
                          ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """ Handles both single instance and batched (inefficiently) """
    # Just anchor volume overlap
    if (vectors_1 is None or vectors_2 is None) and (mask_1 is None or mask_2 is None):
        VAB = overlap_func(anchors_1, anchors_2, alpha)
        VAA = overlap_func(anchors_1, anchors_1, alpha)
        VBB = overlap_func(anchors_2, anchors_2, alpha)
    # Just anchor volume overlap with masking for batching
    elif (vectors_1 is None or vectors_2 is None) and (mask_1 is not None or mask_2 is not None):
        VAB = overlap_func(anchors_1, anchors_2, alpha, mask_1=mask_1, mask_2=mask_2)
        VAA = overlap_func(anchors_1, anchors_1, alpha, mask_1=mask_1, mask_2=mask_1)
        VBB = overlap_func(anchors_2, anchors_2, alpha, mask_1=mask_2, mask_2=mask_2)
    # Anchor and vector volume overlap with masking for batching
    elif mask_1 is not None or mask_2 is not None:
        VAB = overlap_func(anchors_1, anchors_2, vectors_1, vectors_2, alpha, allow_antiparallel, mask_1=mask_1, mask_2=mask_2)
        VAA = overlap_func(anchors_1, anchors_1, vectors_1, vectors_1, alpha, allow_antiparallel, mask_1=mask_1, mask_2=mask_1)
        VBB = overlap_func(anchors_2, anchors_2, vectors_2, vectors_2, alpha, allow_antiparallel, mask_1=mask_2, mask_2=mask_2)
    # Anchor and vector volume overlap for single instance
    else:
        VAB = overlap_func(anchors_1, anchors_2, vectors_1, vectors_2, alpha, allow_antiparallel)
        VAA = overlap_func(anchors_1, anchors_1, vectors_1, vectors_1, alpha, allow_antiparallel)
        VBB = overlap_func(anchors_2, anchors_2, vectors_2, vectors_2, alpha, allow_antiparallel)
    return VAB, VAA, VBB


def _compute_all_overlaps_batch(overlap_func: Callable,
                                cdist_21,
                                cdist_22,
                                cdist_11,
                                alpha: float,
                                vmm_21: Union[torch.Tensor, None] = None,
                                vmm_22: Union[torch.Tensor, None] = None,
                                vmm_11: Union[torch.Tensor, None] = None,
                                allow_antiparallel: bool = False,
                                mask_1: Union[torch.Tensor, None] = None,
                                mask_2: Union[torch.Tensor, None] = None
                                ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """ Only handles batched """
    # Just anchor volume overlap with masking for batching
    if (vmm_21 is None or vmm_22 is None or vmm_11 is None) and (mask_1 is not None or mask_2 is not None):
        VAB = overlap_func(cdist_21, alpha, mask_1=mask_1, mask_2=mask_2)
        VAA = overlap_func(cdist_11, alpha, mask_1=mask_1, mask_2=mask_1)
        VBB = overlap_func(cdist_22, alpha, mask_1=mask_2, mask_2=mask_2)
    # Anchor and vector volume overlap with masking for batching
    elif mask_1 is not None or mask_2 is not None:
        VAB = overlap_func(cdist_21, vmm_21, alpha, allow_antiparallel, mask_1=mask_1, mask_2=mask_2)
        VAA = overlap_func(cdist_11, vmm_11, alpha, allow_antiparallel, mask_1=mask_1, mask_2=mask_1)
        VBB = overlap_func(cdist_22, vmm_22, alpha, allow_antiparallel, mask_1=mask_2, mask_2=mask_2)
    return VAB, VAA, VBB


def tanimoto_func(VAB: torch.Tensor,
                  VAA: torch.Tensor,
                  VBB: torch.Tensor
                  ) -> torch.Tensor:
    """
    Computes Tanimoto similarity.
    Similarity(Tanimoto) = Overlap{1,2} / (Overlap{1,1} + Overlap{2,2} - Overlap{1,2})
    """
    return VAB/(VAA + VBB - VAB)


def tversky_func(VAB: torch.Tensor,
                 VAA: torch.Tensor,
                 VBB: torch.Tensor,
                 sigma: float
                 ) -> torch.Tensor:
    """
    Computes Tversky similarity -> clamped to be max of 1.0.
    sigma: [0,1]

    Similarity(Tversky) = Overlap{1,2} / (sigma*Overlap{1,1} + (1-sigma)*Overlap{2,2})
    """
    return torch.clamp_max(VAB/(sigma*VAA + (1-sigma)*VBB), max=1.0)


def get_vector_volume_overlap_score(ptype_str: str,
                                    ptype_1: torch.Tensor,
                                    ptype_2: torch.Tensor,
                                    anchors_1: torch.Tensor,
                                    anchors_2: torch.Tensor,
                                    vectors_1: torch.Tensor,
                                    vectors_2: torch.Tensor,
                                    allow_antiparallel: bool
                                    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Compute volumentric overlap score with cosine similarity of vectors. Handles batching.
    """
    ptype_str = ptype_str.lower()
    ptype_idx = P_TYPES_LWRCASE.index(ptype_str)
    # single instance
    mask_inds_1 = torch.where(ptype_1 == ptype_idx)[0]
    mask_inds_2 = torch.where(ptype_2 == ptype_idx)[0]
    if len(mask_inds_1) == 0 and len(mask_inds_2) == 0:
        VAB, VAA, VBB = 0., 0., 0.
    elif len(mask_inds_1) == 0:
        VAB, VAA = 0., 0.
        VBB = _compute_fit_overlap(overlap_func=VAB_2nd_order_cosine,
                                    anchors_2=anchors_2[mask_inds_2],
                                    vectors_2=vectors_2[mask_inds_2],
                                    alpha = P_ALPHAS[ptype_str],
                                    allow_antiparallel = allow_antiparallel)
    elif len(mask_inds_2) == 0:
        VAB, VBB = 0., 0.
        VAA = _compute_ref_overlap(overlap_func=VAB_2nd_order_cosine,
                                    anchors_1=anchors_1[mask_inds_1],
                                    vectors_1=vectors_1[mask_inds_1],
                                    alpha = P_ALPHAS[ptype_str],
                                    allow_antiparallel = allow_antiparallel)
    else:
        VAB, VAA, VBB = _compute_all_overlaps(overlap_func=VAB_2nd_order_cosine,
                                                anchors_1=anchors_1[mask_inds_1],
                                                anchors_2=anchors_2[mask_inds_2],
                                                vectors_1=vectors_1[mask_inds_1],
                                                vectors_2=vectors_2[mask_inds_2],
                                                alpha = P_ALPHAS[ptype_str],
                                                allow_antiparallel = allow_antiparallel)
    return VAB, VAA, VBB


def get_vector_volume_overlap_score_batch(ptype_str: str,
                                          ptype_1: torch.Tensor,
                                          ptype_2: torch.Tensor,
                                          cdist_21: torch.Tensor,
                                          cdist_22: torch.Tensor,
                                          cdist_11: torch.Tensor,
                                          vmm_21: torch.Tensor,
                                          vmm_22: torch.Tensor,
                                          vmm_11: torch.Tensor,
                                          allow_antiparallel: bool = False
                                          ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Compute volumentric overlap score with cosine similarity of vectors. Only batching.
    """
    ptype_str = ptype_str.lower()
    ptype_idx = P_TYPES_LWRCASE.index(ptype_str)
    # batched
    mask_1 = torch.where(ptype_1 == ptype_idx, 1, 0)
    mask_2 = torch.where(ptype_2 == ptype_idx, 1, 0)
    VAB, VAA, VBB = _compute_all_overlaps_batch(overlap_func=VAB_2nd_order_cosine_mask_batch,
                                                cdist_21=cdist_21,
                                                cdist_22=cdist_22,
                                                cdist_11=cdist_11,
                                                vmm_21=vmm_21,
                                                vmm_22=vmm_22,
                                                vmm_11=vmm_11,
                                                alpha = P_ALPHAS[ptype_str],
                                                allow_antiparallel = allow_antiparallel,
                                                mask_1 = mask_1,
                                                mask_2 = mask_2)
    return VAB, VAA, VBB


def get_volume_overlap_score(ptype_str: str,
                             ptype_1: torch.Tensor,
                             ptype_2: torch.Tensor,
                             anchors_1: torch.Tensor,
                             anchors_2: torch.Tensor
                             ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Computes volume overlap score single instance.
    """
    ptype_str = ptype_str.lower()
    ptype_idx = P_TYPES_LWRCASE.index(ptype_str)
    # single instance
    mask_inds_1 = torch.where(ptype_1 == ptype_idx)[0]
    mask_inds_2 = torch.where(ptype_2 == ptype_idx)[0]
    if len(mask_inds_1) == 0 and len(mask_inds_2) == 0:
        VAB, VAA, VBB = 0., 0., 0.
    elif len(mask_inds_1) == 0:
        VAB, VAA = 0., 0.
        VBB = _compute_fit_overlap(overlap_func=VAB_2nd_order,
                                    anchors_2=anchors_2[mask_inds_2],
                                    alpha = P_ALPHAS[ptype_str])
    elif len(mask_inds_2) == 0:
        VAB, VBB = 0., 0.
        VAA = _compute_ref_overlap(overlap_func=VAB_2nd_order,
                                    anchors_1=anchors_1[mask_inds_1],
                                    alpha = P_ALPHAS[ptype_str])
    else:
        VAB, VAA, VBB = _compute_all_overlaps(overlap_func=VAB_2nd_order,
                                                anchors_1=anchors_1[mask_inds_1],
                                                anchors_2=anchors_2[mask_inds_2],
                                                alpha = P_ALPHAS[ptype_str])
    return VAB, VAA, VBB


def get_volume_overlap_score_batch(ptype_str: str,
                                   ptype_1: torch.Tensor,
                                   ptype_2: torch.Tensor,
                                   cdist_21: torch.Tensor,
                                   cdist_22: torch.Tensor,
                                   cdist_11: torch.Tensor
                                   ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Compute volumentric overlap score with cosine similarity of vectors. Only batching.
    """
    ptype_str = ptype_str.lower()
    ptype_idx = P_TYPES_LWRCASE.index(ptype_str)
    # batched
    mask_1 = torch.where(ptype_1 == ptype_idx, 1, 0)
    mask_2 = torch.where(ptype_2 == ptype_idx, 1, 0)
    VAB, VAA, VBB = _compute_all_overlaps_batch(overlap_func=VAB_2nd_order_mask_batch,
                                                cdist_21=cdist_21,
                                                cdist_22=cdist_22,
                                                cdist_11=cdist_11,
                                                alpha = P_ALPHAS[ptype_str],
                                                mask_1 = mask_1,
                                                mask_2 = mask_2)
    return VAB, VAA, VBB


def get_volume_overlap_score_extended_points(ptype_str: str,
                                             ptype_1: torch.Tensor,
                                             ptype_2: torch.Tensor,
                                             anchors_1: torch.Tensor,
                                             anchors_2: torch.Tensor,
                                             vectors_1: torch.Tensor,
                                             vectors_2: torch.Tensor,
                                             only_extended: bool = False
                                             ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Score both the anchor and extended point volume overlap instead of a vector similarity.
    """
    ptype_str = ptype_str.lower()
    ptype_idx = P_TYPES_LWRCASE.index(ptype_str)
    # single instance
    if len(ptype_1.shape) == 1 and len(ptype_2.shape) == 1:
        VAB, VAA, VBB = 0., 0., 0.
        mask_inds_1 = torch.where(ptype_1 == ptype_idx)[0]
        mask_inds_2 = torch.where(ptype_2 == ptype_idx)[0]
        if len(mask_inds_1) == 0 and len(mask_inds_2) == 0:
            VAB, VAA, VBB = 0., 0., 0.
        elif len(mask_inds_1) == 0:
            VAB, VAA = 0., 0.
            if not only_extended:
                VBB = _compute_fit_overlap(overlap_func=VAB_2nd_order,
                                           anchors_2=anchors_2[mask_inds_2],
                                           alpha = P_ALPHAS[ptype_str])
            else:
                VBB = 0.
            VBB += _compute_fit_overlap(overlap_func=VAB_2nd_order,
                                        anchors_2=vectors_2[mask_inds_2]+anchors_2[mask_inds_2],
                                        alpha = P_ALPHAS[ptype_str])
        elif len(mask_inds_2) == 0:
            VAB, VBB = 0., 0.
            if not only_extended:
                VAA = _compute_ref_overlap(overlap_func=VAB_2nd_order,
                                           anchors_1=anchors_1[mask_inds_1],
                                           alpha = P_ALPHAS[ptype_str])
            else:
                VAA = 0.
            VAA += _compute_ref_overlap(overlap_func=VAB_2nd_order,
                                        anchors_1=vectors_1[mask_inds_1]+anchors_1[mask_inds_1],
                                        alpha = P_ALPHAS[ptype_str])
        else:
            if not only_extended:
                VAB, VAA, VBB = _compute_all_overlaps(overlap_func=VAB_2nd_order,
                                                      anchors_1=anchors_1[mask_inds_1],
                                                      anchors_2=anchors_2[mask_inds_2],
                                                      alpha = P_ALPHAS[ptype_str])
            VAB_extended, VAA_extended, VBB_extended = _compute_all_overlaps(
                overlap_func=VAB_2nd_order,
                anchors_1=vectors_1[mask_inds_1]+anchors_1[mask_inds_1],
                anchors_2=vectors_2[mask_inds_2]+anchors_2[mask_inds_2],
                alpha = P_ALPHAS[ptype_str]
                )
            VAB += VAB_extended
            VAA += VAA_extended
            VBB += VBB_extended

    # batched
    elif len(ptype_1.shape) == 2 and len(ptype_2.shape) == 2:
        mask_1 = torch.where(ptype_1 == ptype_idx, 1, 0)
        mask_2 = torch.where(ptype_2 == ptype_idx, 1, 0)
        VAB = torch.zeros(ptype_1.shape[0], device=anchors_1.device)
        VAA = torch.zeros(ptype_1.shape[0], device=anchors_1.device)
        VBB = torch.zeros(ptype_1.shape[0], device=anchors_1.device)
        if not only_extended:
            VAB, VAA, VBB = _compute_all_overlaps(overlap_func=VAB_2nd_order_mask,
                                                  anchors_1=anchors_1,
                                                  anchors_2=anchors_2,
                                                  alpha = P_ALPHAS[ptype_str],
                                                  mask_1 = mask_1,
                                                  mask_2 = mask_2
                                                  )
        VAB_extended, VAA_extended, VBB_extended = _compute_all_overlaps(
            overlap_func = VAB_2nd_order_mask,
            anchors_1 = vectors_1+anchors_1,
            anchors_2 = vectors_2+anchors_2,
            alpha = P_ALPHAS[ptype_str],
            mask_1 = mask_1,
            mask_2 = mask_2
            )
        VAB += VAB_extended
        VAA += VAA_extended
        VBB += VBB_extended
    # otherwise
    else:
        raise ValueError(f"Arguments `ptype_1` and `ptype_2` must either be 1D or 2D (batched). Instead these shapes were given: {ptype_1.shape}, {ptype_2.shape}")
    return VAB, VAA, VBB

_SIM_TYPE = Literal['tanimoto', 'tversky', 'tversky_ref', 'tversky_fit']

def get_overlap_pharm(ptype_1: torch.Tensor,
                      ptype_2: torch.Tensor,
                      anchors_1: torch.Tensor,
                      anchors_2: torch.Tensor,
                      vectors_1: torch.Tensor,
                      vectors_2: torch.Tensor,
                      similarity: _SIM_TYPE = 'tanimoto',
                      extended_points: bool = False,
                      only_extended: bool = False
                      ) -> torch.Tensor:
    """
    Compute pharmacophore score.

    Accepts batching, but only if they are the same two molecules (or have the same
    number of features). Specifically used for alignment.

    Parameters
    ----------
    ptype_1 : torch.Tensor
        Indices specifying the pharmacophore type based on order of P_TYPES,
        shape (N,) or (B, N).
    ptype_2 : torch.Tensor
        Indices specifying the pharmacophore type based on order of P_TYPES,
        shape (M,) or (B, M).
    anchors_1 : torch.Tensor
        Coordinates for the anchor points of each pharmacophore of molecule 1,
        shape (N, 3) or (B, N, 3).
    anchors_2 : torch.Tensor
        Coordinates for the anchor points of each pharmacophore of molecule 2,
        shape (M, 3) or (B, M, 3).
    vectors_1 : torch.Tensor
        Relative unit vectors of each pharmacophore of molecule 1,
        shape (N, 3) or (B, N, 3).
    vectors_2 : torch.Tensor
        Relative unit vectors of each pharmacophore of molecule 2,
        shape (M, 3) or (B, M, 3).
    similarity : str, optional
        Specifies what similarity function to use. Options are:
        'tanimoto' (symmetric), 'tversky' (OpenEye's 95% normalization by mol 1),
        'tversky_ref' (Pharao's 100% normalization by mol 1),
        'tversky_fit' (Pharao's 100% normalization by mol 2). Default is 'tanimoto'.
    extended_points : bool, optional
        Whether to score HBA/HBD with gaussian overlaps of extended points.
        Default is ``False``.
    only_extended : bool, optional
        When ``extended_points`` is ``True``, decide whether to only score the
        extended points (ignore anchor overlaps). Default is ``False``.

    Returns
    -------
    torch.Tensor
        Score(s) with shape (1,) or (B,).
    """
    if isinstance(ptype_1, np.ndarray):
        ptype_1 = torch.Tensor(ptype_1)
    if isinstance(anchors_1, np.ndarray):
        anchors_1 = torch.Tensor(anchors_1)
    if isinstance(vectors_1, np.ndarray):
        vectors_1 = torch.Tensor(vectors_1)
    if isinstance(ptype_2, np.ndarray):
        ptype_2 = torch.Tensor(ptype_2)
    if isinstance(anchors_2, np.ndarray):
        anchors_2 = torch.Tensor(anchors_2)
    if isinstance(vectors_2, np.ndarray):
        vectors_2 = torch.Tensor(vectors_2)

    # Similarity scoring
    if similarity.lower() == 'tanimoto':
        similarity_func = tanimoto_func
    elif similarity.lower() == 'tversky':
        similarity_func = partial(tversky_func, sigma=0.95)
    elif similarity.lower() == 'tversky_ref':
        similarity_func = partial(tversky_func, sigma=1.)
    elif similarity.lower() == 'tversky_fit':
        similarity_func = partial(tversky_func, sigma=0.05)
    else:
        raise ValueError('Argument `similarity` must be one of (tanimoto, tversky, tversky_ref, tversky_fit).')

    # Determine if single instance or batched
    if len(ptype_1.shape) == 1 and len(ptype_2.shape) == 1:
        batched = False
        assert anchors_1.shape == (ptype_1.shape[0], 3) and vectors_1.shape == anchors_1.shape, \
            f"Shapes of `anchors_1`, `vectors_1`, and `ptype_1` don't match. Should be (N,3), (N,3), and (N,) but {anchors_1.shape}, {vectors_1.shape}, and {ptype_1.shape} were given."
        assert anchors_2.shape == (ptype_2.shape[0], 3) and vectors_2.shape == anchors_2.shape, \
            f"Shapes of `anchors_2`, `vectors_2`, and `ptype_2` don't match. Should be (N,3), (N,3), and (N,) but {anchors_2.shape}, {vectors_2.shape}, and {ptype_2.shape} were given."

    elif len(ptype_1.shape) == 2 and len(ptype_2.shape) == 2:
        batched = True
        assert anchors_1.shape == (ptype_1.shape + (3,)) and vectors_1.shape == anchors_1.shape, \
            f"Shapes of `anchors_1`, `vectors_1`, and `ptype_1` don't match. Should be (B,N,3), (B,N,3), and (B,N) but {anchors_1.shape}, {vectors_1.shape}, and {ptype_1.shape} were given."
        assert anchors_2.shape == (ptype_2.shape + (3,)) and vectors_2.shape == anchors_2.shape, \
            f"Shapes of `anchors_2`, `vectors_2`, and `ptype_2` don't match. Should be (B,N,3), (B,N,3), and (B,N) but {anchors_2.shape}, {vectors_2.shape}, and {ptype_2.shape} were given."
    else:
        raise ValueError(f"Arguments `ptype_1` and `ptype_2` must either be 1D or 2D (batched). Instead these shapes were given: {ptype_1.shape}, {ptype_2.shape}")

    # Pharmacophores present in the molecules
    ptype_key2ind = {}
    unique_ptypes = torch.concat((torch.unique(ptype_1), torch.unique(ptype_2))).unique()
    for i, ptype_ind in enumerate(unique_ptypes.type(torch.int16)):
        ptype_key2ind[P_TYPES_LWRCASE[ptype_ind.item()]] = i

    # Initialize scores
    if batched:
        overlap = torch.zeros(ptype_1.shape[0]).to(anchors_1.device)
        ref_overlap = torch.zeros(ptype_1.shape[0]).to(anchors_1.device)
        fit_overlap = torch.zeros(ptype_1.shape[0]).to(anchors_1.device)

        vectors_1 = F.normalize(vectors_1, p=2, dim=2)
        vectors_2 = F.normalize(vectors_2, p=2, dim=2)
        cdist_21 = (torch.cdist(anchors_1, anchors_2)**2.0).permute(0,2,1)
        cdist_22 = (torch.cdist(anchors_2, anchors_2)**2.0)
        cdist_11 = (torch.cdist(anchors_1, anchors_1)**2.0)
        vmm_21 = torch.matmul(vectors_1, vectors_2.permute(0,2,1)).permute(0,2,1)
        vmm_22 = torch.matmul(vectors_2, vectors_2.permute(0,2,1))
        vmm_11 = torch.matmul(vectors_1, vectors_1.permute(0,2,1))
    else:
        overlap, ref_overlap, fit_overlap = 0., 0., 0.

    ## Score pharmacophores
    # Hydrophobe
    if 'hydrophobe' in ptype_key2ind:
        if batched:
            VAB, VAA, VBB = get_volume_overlap_score_batch(ptype_str='hydrophobe',
                                                           ptype_1=ptype_1,
                                                           ptype_2=ptype_2,
                                                           cdist_21 = cdist_21,
                                                           cdist_22 = cdist_22,
                                                           cdist_11 = cdist_11)
        else:
            VAB, VAA, VBB = get_volume_overlap_score(ptype_str='hydrophobe',
                                                     ptype_1=ptype_1,
                                                     ptype_2=ptype_2,
                                                     anchors_1=anchors_1,
                                                     anchors_2=anchors_2)
        overlap += VAB
        ref_overlap += VAA
        fit_overlap += VBB

    # Zn
    if 'znbinder' in ptype_key2ind:
        if batched:
            VAB, VAA, VBB = get_volume_overlap_score_batch(ptype_str='znbinder',
                                                           ptype_1=ptype_1,
                                                           ptype_2=ptype_2,
                                                           cdist_21 = cdist_21,
                                                           cdist_22 = cdist_22,
                                                           cdist_11 = cdist_11)
        else:
            VAB, VAA, VBB = get_volume_overlap_score(ptype_str='znbinder',
                                                    ptype_1=ptype_1,
                                                    ptype_2=ptype_2,
                                                    anchors_1=anchors_1,
                                                    anchors_2=anchors_2)
        overlap += VAB
        ref_overlap += VAA
        fit_overlap += VBB

    # Anion
    if 'anion' in ptype_key2ind:
        if batched:
            VAB, VAA, VBB = get_volume_overlap_score_batch(ptype_str='anion',
                                                           ptype_1=ptype_1,
                                                           ptype_2=ptype_2,
                                                           cdist_21 = cdist_21,
                                                           cdist_22 = cdist_22,
                                                           cdist_11 = cdist_11)
        else:
            VAB, VAA, VBB = get_volume_overlap_score(ptype_str='anion',
                                                    ptype_1=ptype_1,
                                                    ptype_2=ptype_2,
                                                    anchors_1=anchors_1,
                                                    anchors_2=anchors_2)
        overlap += VAB
        ref_overlap += VAA
        fit_overlap += VBB

    # Cation
    if 'cation' in ptype_key2ind:
        if batched:
            VAB, VAA, VBB = get_volume_overlap_score_batch(ptype_str='cation',
                                                           ptype_1=ptype_1,
                                                           ptype_2=ptype_2,
                                                           cdist_21 = cdist_21,
                                                           cdist_22 = cdist_22,
                                                           cdist_11 = cdist_11)
        else:
            VAB, VAA, VBB = get_volume_overlap_score(ptype_str='cation',
                                                    ptype_1=ptype_1,
                                                    ptype_2=ptype_2,
                                                    anchors_1=anchors_1,
                                                    anchors_2=anchors_2)
        overlap += VAB
        ref_overlap += VAA
        fit_overlap += VBB

    # Aromatic
    if 'aromatic' in ptype_key2ind:
        if batched:
            VAB, VAA, VBB = get_vector_volume_overlap_score_batch(ptype_str='aromatic',
                                                                  ptype_1=ptype_1,
                                                                  ptype_2=ptype_2,
                                                                  cdist_21 = cdist_21,
                                                                  cdist_22 = cdist_22,
                                                                  cdist_11 = cdist_11,
                                                                  vmm_21 = vmm_21,
                                                                  vmm_22 = vmm_22,
                                                                  vmm_11 = vmm_11,
                                                                  allow_antiparallel=True)
        else:
            VAB, VAA, VBB = get_vector_volume_overlap_score(ptype_str='aromatic',
                                                            ptype_1=ptype_1,
                                                            ptype_2=ptype_2,
                                                            anchors_1=anchors_1,
                                                            anchors_2=anchors_2,
                                                            vectors_1=vectors_1,
                                                            vectors_2=vectors_2,
                                                            allow_antiparallel=True)
        overlap += VAB
        ref_overlap += VAA
        fit_overlap += VBB

    # Acceptor
    if 'acceptor' in ptype_key2ind:
        if extended_points:
            VAB, VAA, VBB = get_volume_overlap_score_extended_points(ptype_str='acceptor',
                                                                     ptype_1=ptype_1,
                                                                     ptype_2=ptype_2,
                                                                     anchors_1=anchors_1,
                                                                     anchors_2=anchors_2,
                                                                     vectors_1=vectors_1,
                                                                     vectors_2=vectors_2,
                                                                     only_extended=only_extended)
        else:
            if batched:
                VAB, VAA, VBB = get_vector_volume_overlap_score_batch(ptype_str='acceptor',
                                                                      ptype_1=ptype_1,
                                                                      ptype_2=ptype_2,
                                                                      cdist_21 = cdist_21,
                                                                      cdist_22 = cdist_22,
                                                                      cdist_11 = cdist_11,
                                                                      vmm_21 = vmm_21,
                                                                      vmm_22 = vmm_22,
                                                                      vmm_11 = vmm_11,
                                                                      allow_antiparallel=False)
            else:
                VAB, VAA, VBB = get_vector_volume_overlap_score(ptype_str='acceptor',
                                                                ptype_1=ptype_1,
                                                                ptype_2=ptype_2,
                                                                anchors_1=anchors_1,
                                                                anchors_2=anchors_2,
                                                                vectors_1=vectors_1,
                                                                vectors_2=vectors_2,
                                                                allow_antiparallel=False)
        overlap += VAB
        ref_overlap += VAA
        fit_overlap += VBB

    # Donors
    if 'donor' in ptype_key2ind:
        if extended_points:
            VAB, VAA, VBB = get_volume_overlap_score_extended_points(ptype_str='donor',
                                                                     ptype_1=ptype_1,
                                                                     ptype_2=ptype_2,
                                                                     anchors_1=anchors_1,
                                                                     anchors_2=anchors_2,
                                                                     vectors_1=vectors_1,
                                                                     vectors_2=vectors_2,
                                                                     only_extended=only_extended)
        else:
            if batched:
                VAB, VAA, VBB = get_vector_volume_overlap_score_batch(ptype_str='donor',
                                                                      ptype_1=ptype_1,
                                                                      ptype_2=ptype_2,
                                                                      cdist_21 = cdist_21,
                                                                      cdist_22 = cdist_22,
                                                                      cdist_11 = cdist_11,
                                                                      vmm_21 = vmm_21,
                                                                      vmm_22 = vmm_22,
                                                                      vmm_11 = vmm_11,
                                                                      allow_antiparallel=False)
            else:
                VAB, VAA, VBB = get_vector_volume_overlap_score(ptype_str='donor',
                                                                ptype_1=ptype_1,
                                                                ptype_2=ptype_2,
                                                                anchors_1=anchors_1,
                                                                anchors_2=anchors_2,
                                                                vectors_1=vectors_1,
                                                                vectors_2=vectors_2,
                                                                allow_antiparallel=False)
        overlap += VAB
        ref_overlap += VAA
        fit_overlap += VBB

    # Halogen
    if 'halogen' in ptype_key2ind:
        if extended_points:
            VAB, VAA, VBB = get_volume_overlap_score_extended_points(ptype_str='halogen',
                                                                     ptype_1=ptype_1,
                                                                     ptype_2=ptype_2,
                                                                     anchors_1=anchors_1,
                                                                     anchors_2=anchors_2,
                                                                     vectors_1=vectors_1,
                                                                     vectors_2=vectors_2,
                                                                     only_extended=only_extended)
        else:
            if batched:
                VAB, VAA, VBB = get_vector_volume_overlap_score_batch(ptype_str='halogen',
                                                                      ptype_1=ptype_1,
                                                                      ptype_2=ptype_2,
                                                                      cdist_21 = cdist_21,
                                                                      cdist_22 = cdist_22,
                                                                      cdist_11 = cdist_11,
                                                                      vmm_21 = vmm_21,
                                                                      vmm_22 = vmm_22,
                                                                      vmm_11 = vmm_11,
                                                                      allow_antiparallel=False)
            else:
                VAB, VAA, VBB = get_vector_volume_overlap_score(ptype_str='halogen',
                                                                ptype_1=ptype_1,
                                                                ptype_2=ptype_2,
                                                                anchors_1=anchors_1,
                                                                anchors_2=anchors_2,
                                                                vectors_1=vectors_1,
                                                                vectors_2=vectors_2,
                                                                allow_antiparallel=False)
        overlap += VAB
        ref_overlap += VAA
        fit_overlap += VBB

    scores = similarity_func(VAB=overlap, VAA=ref_overlap, VBB=fit_overlap)
    return scores


def get_pharm_combo_score(centers_1: torch.Tensor,
                          centers_2: torch.Tensor,
                          ptype_1: torch.Tensor,
                          ptype_2: torch.Tensor,
                          anchors_1: torch.Tensor,
                          anchors_2: torch.Tensor,
                          vectors_1: torch.Tensor,
                          vectors_2: torch.Tensor,
                          alpha: float = 0.81,
                          similarity: str = 'tanimoto',
                          extended_points: bool = False,
                          only_extended: bool = False
                          ) -> torch.Tensor:
    """ Compute a combined shape and pharmacophore score. """
    # Similarity scoring
    if similarity.lower() == 'tanimoto':
        similarity_func = tanimoto_func
    elif similarity.lower() == 'tversky':
        similarity_func = partial(tversky_func, sigma=0.95)
    elif similarity.lower() == 'tversky_ref':
        similarity_func = partial(tversky_func, sigma=1.)
    elif similarity.lower() == 'tversky_fit':
        similarity_func = partial(tversky_func, sigma=0.05)
    else:
        raise ValueError('Argument `similarity` must be one of (tanimoto, tversky, tversky_ref, tversky_fit).')

    # Pharmacophore scoring
    pharm_score = get_overlap_pharm(ptype_1=ptype_1,
                                    ptype_2=ptype_2,
                                    anchors_1=anchors_1,
                                    anchors_2=anchors_2,
                                    vectors_1=vectors_1,
                                    vectors_2=vectors_2,
                                    similarity=similarity,
                                    extended_points=extended_points,
                                    only_extended=only_extended)

    # Shape scoring
    VAB = VAB_2nd_order(centers_1=centers_1,
                        centers_2=centers_2,
                        alpha=alpha)
    VAA = VAB_2nd_order(centers_1=centers_1,
                        centers_2=centers_1,
                        alpha=alpha)
    VBB = VAB_2nd_order(centers_1=centers_2,
                        centers_2=centers_2,
                        alpha=alpha)
    shape_score = similarity_func(VAB=VAB,
                                  VAA=VAA,
                                  VBB=VBB)

    score = (pharm_score + shape_score)/2
    return score



"""
Docking benchmark:
Take 10k test molecules from MOSES and dock them against each target.
"""
import argparse
from pathlib import Path

import pickle

from rdkit import Chem

from shepherd_score.evaluations.utils.convert_data import get_smiles_from_atom_pos
from shepherd_score.evaluations.docking import docking_target_info, DockingEvalPipeline

experimental_pdb_order = ['3ny8', '7l11', '3eml', '4unn', '4rlu', '1iep', '5mo4']

docking_screen_pdb_order = ['1iep', '3eml', '3ny8', '4rlu', '4unn', '5mo4', '7l11']

pdb_bound_poses_order = {
    'Cc1ccc(NC(=O)c2ccc(CN3CCN(C)CC3)cc2)cc1Nc1nccc(-c2cccnc2)n1': '1iep',
    'Nc1nc(NCCc2ccc(O)cc2)nc2nc(-c3ccco3)nn12' : '3eml',
    'Cc1ccc(OC[C@@H](O)[C@H](C)NC(C)C)c2c1CCC2' : '3ny8',
    'O=C(C=Cc1ccc(O)cc1)c1ccc(O)cc1O': '4rlu',
    'COc1cccc(-c2cc(-c3ccc(C(=O)O)cc3)c(C#N)c(=O)[nH]2)c1': '4unn',
    'O=C(Nc1ccc(OC(F)(F)Cl)cc1)c1cnc(N2CC[C@@H](O)C2)c(-c2ccn[nH]2)c1': '5mo4',
    'CCCOc1cc(Cl)cc(-c2cc(-c3ccccc3C#N)cn(-c3cccnc3)c2=O)c1': '7l11'
}


def docking_experiment_evaluation(pdb_id: str, sample_idx, file_path, save_path_dir: Path, my_task_id, num_tasks):
    """
    Run the docking evaluation pipeline.
    """
    docking_pipe = DockingEvalPipeline(pdb_id=pdb_id,
                                       num_processes=4,
                                       verbose=0,
                                       path_to_bin='')

    with open(file_path, 'rb') as f:
        samples = pickle.load(f)

    out_dir = save_path_dir / f'{pdb_id}_{sample_idx}'
    out_dir.mkdir(parents=True, exist_ok=True)
    assert out_dir.is_dir()
    out_file = out_dir / f'docking_eval_{my_task_id}.pkl'

    # Load all the generated samples as SMILES
    smiles_list = []
    for sample in samples[-1]:
        smiles_list.append(get_smiles_from_atom_pos(sample['x1']['atoms'], sample['x1']['positions']))

    this_smiles_ls = smiles_list[my_task_id:len(smiles_list):num_tasks]

    print('\n\n\nBEGIN DOCKING EVAL')
    docking_pipe.evaluate(this_smiles_ls, exhaustiveness=32, n_poses=1, protonate=False,
                          save_poses_dir_path=None, verbose=True)

    energies = [float(e) for e in docking_pipe.energies]
    smiles = docking_pipe.smiles
    buffer = docking_pipe.buffer

    with open(out_file, 'wb') as f:
        pickle.dump((pdb_id, buffer, smiles, energies), f, protocol=pickle.HIGHEST_PROTOCOL)
    print(f'Saved to {out_file}')


if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--load-dir-path', type=str, help='Path to directory to load files.')
    parser.add_argument('--task-id', type=int, help='Task ID.')
    parser.add_argument('--num-tasks', type=int, help='Number of tasks.')
    parser.add_argument('--sample-idx', type=int, required=True, help='Index used to load the file. [0,6]')
    args = parser.parse_args()
    print(args)

    load_dir = Path(args.load_dir_path)
    if not load_dir.is_dir():
        raise ValueError('Provided --load-dir-path is not a directory.')
    my_task_id = int(args.task_id)
    num_tasks = int(args.num_tasks)
    sample_idx = int(args.sample_idx)

    # Experimental conditioning
    working_dir = load_dir / 'PDB_analogues'
    pdb_id = experimental_pdb_order[sample_idx]

    file_path_experimental = working_dir / f'samples_{sample_idx}.pickle'
    print(f'Loading from {file_path_experimental}')
    docking_experiment_evaluation(pdb_id=pdb_id,
                                  sample_idx=sample_idx,
                                  file_path=file_path_experimental,
                                  save_path_dir=working_dir,
                                  my_task_id=my_task_id,
                                  num_tasks=num_tasks)

    # Docking screen conditioning
    working_dir = load_dir
    file_path_docking_screen = working_dir / f'samples_{sample_idx}.pickle'
    if 'PDB_analogues_pose' in working_dir.stem:
        with open(file_path_docking_screen, 'rb') as f:
            samples = pickle.load(f)
        ref_smiles = Chem.MolToSmiles(Chem.MolFromMolBlock(samples[0], removeHs=True))
        pdb_id = pdb_bound_poses_order[ref_smiles]
    else:
        pdb_id = docking_screen_pdb_order[sample_idx]


    print(f'Loading from {file_path_docking_screen}')
    print(pdb_id)
    docking_experiment_evaluation(pdb_id=pdb_id,
                                  sample_idx=sample_idx,
                                  file_path=file_path_docking_screen,
                                  save_path_dir=working_dir,
                                  my_task_id=my_task_id,
                                  num_tasks=num_tasks)
